from __future__ import annotations
from dataclasses import dataclass
from enum import Enum
from arcgis.geometry import (
    Point,
    Polygon,
    Polyline,
    Envelope,
    MultiPoint,
    SpatialReference,
)


class SpatialRelationship(Enum):
    """
    ==================  ===============================================================================
    **Parameter**        **Description**
    ------------------  -------------------------------------------------------------------------------
    INTERSECTS          Query Geometry Intersects Target Geometry.
    ------------------  -------------------------------------------------------------------------------
    ENVELOPEINTERSECTS  Envelope of Query Geometry Intersects Envelope of Target Geometry.
    ------------------  -------------------------------------------------------------------------------
    INDEXINTERSECTS     Query Geometry Intersects Index entry for Target Geometry (Primary Index Filter).
    ------------------  -------------------------------------------------------------------------------
    TOUCHES             Query Geometry Touches Target Geometry.
    ------------------  -------------------------------------------------------------------------------
    OVERLAPS            Query Geometry Overlaps Target Geometry.
    ------------------  -------------------------------------------------------------------------------
    CROSSES             Query Geometry Crosses Target Geometry.
    ------------------  -------------------------------------------------------------------------------
    WITHIN              Query Geometry is Within Target Geometry.
    ------------------  -------------------------------------------------------------------------------
    CONTAINS            Query Geometry Contains Target Geometry.
    ==================  ===============================================================================
    """

    INTERSECTS = "esriSpatialRelIntersects"
    CONTAINS = "esriSpatialRelContains"
    CROSSES = "esriSpatialRelCrosses"
    ENVELOPEINTERSECTS = "esriSpatialRelEnvelopeIntersects"
    INDEXINTERSECTS = "esriSpatialRelIndexIntersects"
    OVERLAPS = "esriSpatialRelOverlaps"
    TOUCHES = "esriSpatialRelTouches"
    WITHIN = "esriSpatialRelWithin"


@dataclass
class SpatialFilter:
    """Creates a spatial filter that can be used in query and create view operations."""

    geometry: Point | Polygon | Polyline | Envelope | MultiPoint
    spatial_rel: SpatialRelationship = SpatialRelationship.INTERSECTS
    sr: SpatialReference | None = None

    def as_json(self):
        """Converts the SpatialFilter to a JSON representation"""
        gt = {
            "point": "Point",
            "multipoint": "Multipoint",
            "polygon": "Polygon",
            "polyline": "Polyline",
            "envelope": "Envelope",
        }
        spatial_filter = {
            "geometry": self.geometry,
            "geometryType": "esriGeometry" + gt[str(self.geometry.type).lower()],
            "spatialRel": self.spatial_rel.value,
        }

        if self.sr is None:
            if "spatialReference" in self.geometry:
                sr = self.geometry["spatialReference"]

        else:
            spatial_filter["inSR"] = sr
        return spatial_filter



from __future__ import annotations
from cachetools import cached, TTLCache
import lxml.html
from urllib.parse import urlunparse, quote, parse_qsl, parse_qs, urlparse
from functools import lru_cache
from typing import Any
from getpass import getpass
from requests.auth import AuthBase
from requests_oauthlib import OAuth2Session
from requests.cookies import extract_cookies_to_jar

from ._schain import SupportMultiAuth
from ..tools._lazy import LazyLoader
from ..tools import parse_url, assemble_url
from .._error import ArcGISLoginError

warnings = LazyLoader("warnings")
re = LazyLoader("re")
json = LazyLoader("json")
threading = LazyLoader("threading")
_dt = LazyLoader("datetime")
requests = LazyLoader("requests")
requests_oauthlib = LazyLoader("requests_oauthlib")
warnings = LazyLoader("warnings")

_MSG = """

You need to a security question by integer:

1. What city were you born in?
2. What was your high school mascot?
3. What is your mother's maiden name?
4. What was the make of your first car?
5. What high school did you go to?
6. What is the last name of your best friend?
7. What is the middle name of your youngest sibling?
8. What is the name of the street on which you grew up?
9. What is the name of your favorite fictional character?
10. What is the name of your favorite pet?
11. What is the name of your favorite restaurant?
12. What is the title of your favorite book?
13. What is your dream job?
14. Where did you go on your first date?
"""


# -------------------------------------------------------------------------
@lru_cache(maxsize=255)
def _token_url_validator(
    url: str,
    session: "EsriSession",
    verify: bool = False,
    proxies: frozenset = None,
) -> str:
    """validates the token url from the give URL"""
    parts = ["/info", "/rest/info", "/sharing/rest/info"]
    params = {"f": "json"}
    if proxies:
        proxies = dict(proxies)
    parsed_url = _parse_arcgis_url(url=url)
    token_url = None  # parsed_url + "/sharing/rest/generateToken"
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        for pt in parts:
            try:
                resp = session.get(
                    f"{parsed_url}{pt}?f=json",
                    proxies=proxies,
                    verify=verify,
                )  # need to include proxies, verify parameter
                token_url = resp.json()["authInfo"]["tokenServicesUrl"]
                if token_url:
                    break
                del pt
            except Exception as e:
                pass
    return token_url


# -------------------------------------------------------------------------
@lru_cache(maxsize=255)
def _parse_arcgis_url(url: str) -> str:
    """
    Returns a valid ArcGIS Online or ArcGIS Enterprise base URL

    :returns: str
    """
    if url is None or url.lower().find("www.arcgis.com") > -1:
        return "https://www.arcgis.com"
    parsed = parse_url(url)
    if len(parsed.path) <= 1:
        return f"{parsed.scheme}://{parsed.netloc}"
    else:
        idx = parsed.path.lower().find("sharing/rest")
        if idx != -1:
            wa = parsed.path[1:idx]
            wa = wa.replace("/", "")
        else:
            wa = parsed.path[1:].split("/")
            wa = wa[0].replace("/", "")
        if len(wa) == 0:
            return f"{parsed.scheme}://{parsed.netloc}"
        return f"{parsed.scheme}://{parsed.netloc}/{wa}"


###########################################################################
class ArcGISServerAuth(AuthBase, SupportMultiAuth):
    """
    Performs the ArcGIS Server (ags) Authentication for a given request.
    """

    _arcpy = None
    _referer = None
    _invalid_token_urls = None
    _401_counters = None
    _ags = None

    # ----------------------------------------------------------------------
    def __init__(self, ags_file: str, legacy: bool = False):
        try:
            self._arcpy = LazyLoader("arcpy", strict=True)
            self.legacy = legacy
            self._invalid_token_urls = set()
            self._401_counters = dict()
            self._ags = ags_file
        except:
            raise

    # ----------------------------------------------------------------------
    def __str__(self):
        return f"<{self.__class__.__name__}, token=.....>"

    # ----------------------------------------------------------------------
    def __repr__(self):
        return f"<{self.__class__.__name__}, token=.....>"

    # ----------------------------------------------------------------------
    @property
    def token(self):
        """obtains the login token"""
        return self._ags_token()

    @lru_cache(maxsize=255)
    def _read_ags_file(self, ags_file: str) -> dict[str, Any]:
        """reads the ags file into cache"""
        return self._arcpy.gp.getStandaloneServerToken(self._ags) or {}

    # ----------------------------------------------------------------------
    @lru_cache(maxsize=255)
    def _url(self, ags_file) -> str:
        if self._arcpy:
            resp = self._read_ags_file(ags_file)
            return _parse_arcgis_url(resp.get("serverUrl")) + "/rest/services"
        else:
            raise Exception("ArcPy not found, please install arcpy")

    # ----------------------------------------------------------------------
    @property
    def url(self) -> str:
        """gets the token for various products"""
        if self._arcpy:
            return self._url(ags_file=self._ags)

        else:
            raise Exception("ArcPy not found, please install arcpy")

    # ----------------------------------------------------------------------
    def _ags_token(self):
        """gets the token for various products"""
        if self._arcpy:
            resp = self._read_ags_file(self._ags)
            if resp:
                if "referer" in resp:
                    self._referer = resp["referer"]
                if "token" in resp:
                    return resp["token"]
                else:
                    raise Exception("Could not generate token.")
            else:
                raise Exception(
                    (
                        "Could not login using Pro authencation."
                        "Please verify in Pro that you are logged in."
                    )
                )
        else:
            raise Exception("ArcPy not found, please install arcpy")

    # ----------------------------------------------------------------------
    def handle_40x(self, r, **kwargs):
        """Handles Case where token is invalid"""
        parsed = parse_url(r.url)
        server_url = assemble_url(parsed)
        if (r.status_code < 500 and r.status_code > 399) and str(r.text).lower().find(
            "invalid token"
        ) > -1:
            # Recreate the request without the token
            #
            parsed = parse_url(r.url)

            self._invalid_token_urls.add(server_url)
            r.content
            r.raw.release_conn()
            r.request.headers.pop("X-Esri-Authorization", None)
            _r = r.connection.send(r.request, **kwargs)
            _r.headers["referer"] = self._referer or "http"
            _r.history.append(r)
            return _r
        elif str(r.text).lower().find("invalid token") > -1:
            # Recreate the request without the token
            #
            parsed = parse_url(r.url)
            self._invalid_token_urls.add(server_url)
            r.content
            r.raw.release_conn()
            r.request.headers.pop("X-Esri-Authorization", None)
            _r = r.connection.send(r.request, **kwargs)
            _r.headers["referer"] = self._referer or "http"
            _r.history.append(r)
            return _r
        return r

    # ----------------------------------------------------------------------
    def __call__(self, r):
        """Handles the Token Authorization Logic"""
        if self._invalid_token_urls is None:
            self._invalid_token_urls = set()
        parsed = parse_url(r.url)
        server_url = assemble_url(parsed)
        if not server_url in self._invalid_token_urls:
            r.register_hook("response", self.handle_40x)
            if self.legacy == False and self.token:
                r.headers["X-Esri-Authorization"] = f"Bearer {self.token}"
                r.headers["referer"] = self._referer or ""
            elif self.legacy == False and not self.token:
                r.headers["referer"] = self._referer or ""
            elif self.legacy and r.method == "GET" and self.token:
                r.prepare_url(url=r.url, params={"token": self.token})
                r.headers["referer"] = self._referer or ""
            elif self.legacy and r.method == "GET" and not self.token:
                r.headers["referer"] = self._referer or ""
            elif self.legacy and r.method == "POST" and self.token:
                data = parse_qs(r.body)
                data["token"] = self.token
                r.prepare_body(data, None, None)
                r.headers["referer"] = self._referer or ""
            elif self.legacy and r.method == "POST" and not self.token:
                data = parse_qs(r.body)
                r.prepare_body(data, None, None)
                r.headers["referer"] = self._referer or ""
            else:
                raise Exception(
                    "Only POST and GET are supported with legacy methods of authentication."
                )
            try:
                self.pos = r.body.tell()
            except AttributeError:
                self.pos = None

            return r
        else:
            r.headers.pop("X-Esri-Authorization", None)
        return r


###########################################################################
class ArcGISProAuth(AuthBase, SupportMultiAuth):
    """
    Performs the ArcGIS Pro Authentication for a given request.
    """

    _arcpy = None
    _referer = None
    _invalid_token_urls = None
    _401_counters = None

    def __init__(self, legacy: bool = False):
        try:
            self._arcpy = LazyLoader("arcpy", strict=True)
            self.legacy = legacy
            self._invalid_token_urls = set()
            self._401_counters = dict()
        except:
            raise

    # ----------------------------------------------------------------------
    def __str__(self):
        return f"<{self.__class__.__name__}, token=.....>"

    # ----------------------------------------------------------------------
    def __repr__(self):
        return f"<{self.__class__.__name__}, token=.....>"

    # ----------------------------------------------------------------------
    @property
    def token(self):
        """obtains the login token"""
        return self._pro_token()

    # ----------------------------------------------------------------------
    def _pro_token(self):
        """gets the token for various products"""
        if self._arcpy:
            resp = self._arcpy.GetSigninToken()
            if resp:
                if "referer" in resp:
                    self._referer = resp["referer"]
                if "token" in resp:
                    return resp["token"]
                else:
                    raise Exception("Could not generate token.")
            else:
                raise Exception(
                    (
                        "Could not login using Pro authencation."
                        "Please verify in Pro that you are logged in."
                    )
                )
        else:
            raise Exception("ArcPy not found, please install arcpy")

    # ----------------------------------------------------------------------
    def handle_40x(self, r, **kwargs):
        """Handles Case where token is invalid"""
        parsed = parse_url(r.url)
        server_url = assemble_url(parsed)
        if (r.status_code < 500 and r.status_code > 399) and str(r.text).lower().find(
            "invalid token"
        ) > -1:
            # Recreate the request without the token
            #
            parsed = parse_url(r.url)

            self._invalid_token_urls.add(server_url)
            r.content
            r.raw.release_conn()
            r.request.headers.pop("X-Esri-Authorization", None)
            _r = r.connection.send(r.request, **kwargs)
            _r.headers["referer"] = self._referer or "http"
            _r.history.append(r)
            return _r
        elif str(r.text).lower().find("invalid token") > -1:
            # Recreate the request without the token
            #
            parsed = parse_url(r.url)
            self._invalid_token_urls.add(server_url)
            r.content
            r.raw.release_conn()
            r.request.headers.pop("X-Esri-Authorization", None)
            _r = r.connection.send(r.request, **kwargs)
            _r.headers["referer"] = self._referer or "http"
            _r.history.append(r)
            return _r
        return r

    # ----------------------------------------------------------------------
    def __call__(self, r):
        """Handles the Token Authorization Logic"""
        if self._invalid_token_urls is None:
            self._invalid_token_urls = set()
        parsed = parse_url(r.url)
        server_url = assemble_url(parsed)
        if not server_url in self._invalid_token_urls:
            r.register_hook("response", self.handle_40x)
            if self.legacy == False:
                r.headers["X-Esri-Authorization"] = f"Bearer {self.token}"
                r.headers["referer"] = self._referer or ""
            elif self.legacy and r.method == "GET":
                r.prepare_url(url=r.url, params={"token": self.token})
                r.headers["referer"] = self._referer or ""
            elif self.legacy and r.method == "POST":
                data = parse_qs(r.body)
                data["token"] = self.token
                r.prepare_body(data, None, None)
                r.headers["referer"] = self._referer or ""
            else:
                raise Exception(
                    "Only POST and GET are supported with legacy methods of authentication."
                )
            try:
                self.pos = r.body.tell()
            except AttributeError:
                self.pos = None

            return r
        else:
            r.headers.pop("X-Esri-Authorization", None)
        return r


###########################################################################
class EsriBuiltInAuth(AuthBase, SupportMultiAuth):
    """
    Performs the BUILT-IN Login Authorization for ArcGIS Online and Enterprise
    """

    _params = None
    _auth_token = None
    _authorization_url = None
    _state = None
    _auto_refresh_extra_params = None
    _clientid = None
    _username = None
    _password = None
    _re_expiration = None
    _create_time = None
    _expiration_time = None
    _oauth_info = None
    _referer = None
    _no_go_token = None
    _response_type = None
    _client = None
    _expiration = None

    # ----------------------------------------------------------------------
    def __init__(
        self,
        url: str,
        username: str,
        password: str,
        expiration: int = None,
        legacy: bool = False,
        verify_cert: bool = True,
        referer: str = None,
        **kwargs,
    ):
        """init"""
        from requests.packages import urllib3

        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
        self._expiration = expiration or 20160
        self._response_type = kwargs.pop("response_type", "token")
        if self._response_type == "token":
            self._clientid = "pythonapi"  # "arcgisonline"
        else:
            self._clientid = kwargs.get("clientid", "pythonapi")  # "arcgispro"
        if self._response_type == "token":
            from oauthlib.oauth2 import MobileApplicationClient

            self._client = MobileApplicationClient(client_id=self._clientid)

        if referer is None:
            self._referer = ""
        url = _parse_arcgis_url(url=url)
        self.legacy = legacy
        self._verify_cert = verify_cert
        self.proxies = kwargs.pop("proxies", {})
        self._base_url = url
        self._auth_url = f"{url}/sharing/rest/oauth2/authorize"
        self._token_url = f"{url}/sharing/rest/oauth2/token"
        self._signin_url = f"{url}/sharing/oauth2/signin"
        self._reset_password_url = f"{url}/sharing/oauth2/resetPassword"
        self._update_profile_url = f"{url}/sharing/oauth2/updateUserProfile"
        self._mfa_url = f"{url}/sharing/oauth2/mfa"
        self._mfa_code = kwargs.pop("mfa_code", None)
        self._no_go_token = set()
        self._username = username
        self._password = password
        self._referer = referer or ""

        self._re_expressions = {
            "step-1a": re.compile("var oAuthInfo = ({.*?});", re.DOTALL),
            "step-1b": re.compile("var oAuthInfo = ({.*?})", re.DOTALL),
            "step-1c": re.compile("var\s+(\w+)\s*=\s*({.*?})", re.DOTALL),
            "step-2": re.compile(r"<title>SUCCESS code=(.*?)</title>", re.DOTALL),
            "password_reset": re.compile(r"{.*\:.*}"),
        }
        self._params = {"expiration": expiration or 1440}

        self._auto_refresh_extra_params = {"client_id": self._clientid}

    # ----------------------------------------------------------------------
    def __str__(self):
        return f"<{self.__class__.__name__}, token=.....>"

    # ----------------------------------------------------------------------
    def __repr__(self):
        return f"<{self.__class__.__name__}, token=.....>"

    # ----------------------------------------------------------------------
    def suspend(self) -> bool:
        """
        Invalidates the login and checks any licenses back in

        :return: Bool.  If True, the licenses is checked back in successfully,
                        False, the process failed.
        """

        if self._auth_token and "refresh_token" in self._auth_token:
            params = {
                "grant_type": "suspend_session",
                "client_id": self._clientid,  # "arcgispro",
                "refresh_token": self._auth_token["refresh_token"],
                "f": "json",
            }
        else:
            return True
        try:
            return (
                requests.post(
                    self._token_url,
                    data=params,
                    verify=self._verify_cert,
                    proxies=self.proxies,
                )
                .json()
                .get("success", False)
            )
        except:
            return False

    def _init_response_type_token(self):
        """"""
        import requests_oauthlib

        redirect_uri = f"https://{parse_url(self._auth_url).netloc}"  # "urn:ietf:wg:oauth:2.0:oob" does not work for MobileApplicationClient

        session = requests_oauthlib.OAuth2Session(
            self._clientid,
            client=self._client,
            redirect_uri=redirect_uri,
        )
        auth_url, state = session.authorization_url(
            self._auth_url,
            expiration=self._expiration,
            **{
                "allow_verification": "false",
                "style": "dark",
                "locale": "en-US",
            },
        )
        auth_response = requests.get(
            url=auth_url,
            proxies=self.proxies,
            verify=self._verify_cert,
        ).text
        if "oauth_state" in auth_response:
            oauth_state = auth_response.split('"oauth_state":"')[1].split('"')[0]
        else:
            raise Exception("Unable to generate oauth token")
        params = {
            "expiration": self._expiration,
            "oauth_state": oauth_state,
            "username": self._username,
            "password": self._password,
        }

        response = requests.post(
            self._signin_url,
            data=params,
            allow_redirects=False,
            proxies=self.proxies,
            verify=self._verify_cert,
        )
        #
        # After authenticating, ArcGIS Online/Enterprise can prompt for a
        # terms and conditions acceptance.
        #
        if response.text.find("OAUTH_0015") > -1:
            raise ArcGISLoginError()
        callback_url = response.headers["location"]
        if callback_url.find("acceptTermsAndConditions") > -1:
            parsed = parse_url(response.headers["location"])
            oauth_state = parse_qs(parsed.query)["oauth_state"][0]

            url = f"{parsed.scheme}://{parsed.netloc}{parsed.path}"
            params = {
                "oauth_state": oauth_state,
                "acceptTermsAndConditions": True,
            }
            response = requests.post(
                url,
                params,
                verify=self._verify_cert,
                allow_redirects=False,
                proxies=self.proxies,
            )
            callback_url = response.headers["location"]
        self._expiration_time = _dt.datetime.now() + _dt.timedelta(seconds=1440)

        # Now we extract the token from the URL to make use of it.
        self._auth_token = session.token_from_fragment(callback_url)
        if "expires_at" in self._auth_token:
            self._expiration_time = _dt.datetime.fromtimestamp(
                self._auth_token["expires_at"]
            )
        elif "expires" in self._auth_token:
            self._expiration_time = _dt.datetime.now() + _dt.timedelta(
                seconds=self._auth_token["expiration"]
            )
        else:
            self._expiration_time = _dt.datetime.now() + _dt.timedelta(seconds=300)

    # ----------------------------------------------------------------------
    def _init_token_auth_handshake(self):
        """perform initial handshake"""
        redirect_uri = "urn:ietf:wg:oauth:2.0:oob"
        self._oauth = OAuth2Session(self._clientid, redirect_uri=redirect_uri)
        authorization_url, state = self._oauth.authorization_url(
            self._auth_url,
            expiration=20160,
            **{
                "allow_verification": "false",
                "style": "dark",
                "locale": "en-US",
            },
        )
        self._authorization_url = authorization_url
        self._state = state
        content = requests.get(
            self._authorization_url,
            verify=self._verify_cert,
            proxies=self.proxies,
        ).text
        if content.find("Error: Invalid client_id") > -1:
            self._clientid = "arcgisonline"
            from oauthlib.oauth2 import MobileApplicationClient

            self._client = MobileApplicationClient(client_id=self._clientid)
            self._init_response_type_token()
            return
        oauth_info = None
        pattern = self._re_expressions["step-1a"]
        if len(pattern.findall(content)) == 0:
            pattern = self._re_expressions["step-1b"]
        soup = lxml.html.fromstring(content)
        for script in soup.xpath("//script/text()"):
            script_code = str(script).strip()
            matches = pattern.search(script_code)
            if not matches is None:
                js_object = matches.groups()[0]
                try:
                    oauth_info = json.loads(js_object)
                except:
                    oauth_info = json.loads(js_object + "}")
                break
        if oauth_info:
            oauth_state = oauth_info["oauth_state"]
        else:
            raise ArcGISLoginError(
                "Could not login. Please ensure you have valid credentials and set your security login question."
            )

        signin_params = {
            "oauth_state": oauth_state,
            "authorize": True,
            "username": self._username,
            "password": self._password,
        }
        signin_resp = requests.post(
            self._signin_url,
            signin_params,
            verify=self._verify_cert,
            allow_redirects=True,
            proxies=self.proxies,
        )
        matches = pattern.findall(signin_resp.text)
        if len(matches) > 0:
            try:
                sign_json = json.loads(matches[0].strip())
            except:
                sign_json = json.loads(matches[0].strip() + "}")
        else:
            sign_json = {}
        if "messages" in sign_json and (
            any(
                [
                    msg.lower().find("invalid username or password") > -1
                    for msg in sign_json["messages"]
                ]
            )
            or any(
                [
                    msg.find("Too many invalid attempts. Please try again later.") > -1
                    for msg in sign_json["messages"]
                ]
            )
        ):
            raise ValueError(",".join(sign_json["messages"]))
        elif signin_resp.url.lower().find("accepttermsandconditions") > -1:
            parsed = parse_url(signin_resp.url)
            oauth_state = parse_qs(signin_resp.url.split("?")[-1])["oauth_state"][0]
            url = f"{parsed.scheme}://{parsed.netloc}{parsed.path}"
            params = {
                "oauth_state": oauth_state,
                "acceptTermsAndConditions": True,
            }
            signin_resp = requests.post(
                url,
                params,
                verify=self._verify_cert,
                allow_redirects=True,
                proxies=self.proxies,
            )
            resp_text = signin_resp.text
            exp = r"<title>SUCCESS code=(.*?)</title>"
            pattern = self._re_expressions["step-2"]
            code = pattern.findall(resp_text)[0]
        elif signin_resp.url.lower().find("updateuserprofile") > -1:
            # raise Exception(
            # (
            # "This is your first time logging in and you are required"
            # " to setup a new password manually before logging in."
            # )
            # )

            print(
                "This is your first time logging in and you are required to setup a new password."
            )
            new_password = getpass(prompt="Please Enter a New Password: ")
            assert self._password != new_password

            resp_text = signin_resp.text
            pattern = self._re_expressions["password_reset"]
            oauth_state = json.loads(pattern.findall(resp_text)[0].replace(" ", ""))[
                "oauth_state"
            ]
            params = {
                "password": self._password,
                "newPassword": new_password,
                "newPassword2": new_password,
                "oauth_state": oauth_state,
                "f": "json",
            }
            resp = requests.post(
                self._reset_password_url,
                params,
                verify=self._verify_cert,
                allow_redirects=True,
                proxies=self.proxies,
            )
            self._password = new_password
            oauth_state = json.loads(pattern.findall(resp_text)[0].replace(" ", ""))[
                "oauth_state"
            ]
            print(_MSG)
            question = int(getpass(prompt="Select a question by integer: "))
            answer = getpass("Answer to the question: ")
            params = {
                "f": "json",
                "securityQuestionIdx": question,
                "securityAnswer": answer,
                "oauth_state": oauth_state,
            }
            resp = requests.post(
                self._update_profile_url,
                params,
                verify=self._verify_cert,
                allow_redirects=True,
                proxies=self.proxies,
            )
            self._init_token_auth_handshake()
        elif signin_resp.url.lower().find("/mfa") > -1:
            from ..tools._util import mfa_otp

            mfa_code = self._mfa_code
            if self._mfa_code is None:
                verify_code = input("Please input your 2FA code: ")
            else:
                verify_code = mfa_otp(mfa_code)

            oauth_info = None
            pattern = self._re_expressions["step-1a"]
            if len(pattern.findall(content)) == 0:
                pattern = self._re_expressions["step-1b"]
            soup = lxml.html.fromstring(signin_resp.text)
            for script in soup.xpath("//script/text()"):
                script_code = str(script).strip()
                matches = pattern.search(script_code)
                if not matches is None:
                    js_object = matches.groups()[0]
                    try:
                        oauth_info = json.loads(js_object)
                    except:
                        oauth_info = json.loads(js_object + "}")
                    break
            mfa_params: dict[str, str] = {
                "oauth_state": oauth_info["oauth_state"],
                "authResponse": "",
                "totp": "",
                "mfa_code": verify_code,
                "recovery_code": "",
            }

            resp = requests.post(
                self._mfa_url,
                data=mfa_params,
                verify=self._verify_cert,
                allow_redirects=False,
                proxies=self.proxies,
            )
            resp_text = requests.get(resp.headers["location"]).text
            exp = r"<title>SUCCESS code=(.*?)</title>"
            pattern = self._re_expressions["step-2"]
            code = pattern.findall(resp_text)[0]
            self._auth_token = self._oauth.fetch_token(
                token_url=self._token_url,
                code=code,
                verify=self._verify_cert,
                include_client_id=True,
                proxies=self.proxies,
                **{"expiration": 20160},
            )
        else:
            resp_text = signin_resp.text
            exp = r"<title>SUCCESS code=(.*?)</title>"
            pattern = self._re_expressions["step-2"]
            code = pattern.findall(resp_text)[0]
        self._auth_token = self._oauth.fetch_token(
            token_url=self._token_url,
            code=code,
            verify=self._verify_cert,
            include_client_id=True,
            proxies=self.proxies,
            **{"expiration": 20160},
        )
        if "expires_at" in self._auth_token:
            self._expiration_time = _dt.datetime.fromtimestamp(
                self._auth_token["expires_at"]
            )
        elif "expires" in self._auth_token:
            self._expiration_time = _dt.datetime.now() + _dt.timedelta(
                seconds=self._auth_token["expiration"]
            )
        else:
            self._expiration_time = _dt.datetime.now() + _dt.timedelta(seconds=300)

    # ----------------------------------------------------------------------
    @property
    def token(self):
        """obtains the login token"""
        try:
            if self._auth_token:
                if (
                    _dt.datetime.now() + _dt.timedelta(minutes=5)
                ) >= self._expiration_time:
                    self._refresh()
                return self._auth_token["access_token"]
            else:
                if self._response_type == "token":
                    self._init_response_type_token()
                else:
                    self._init_token_auth_handshake()
                return self.token
        except:
            self._auth_token = None
            self._init_token_auth_handshake()
            if self._auth_token:
                return self.token
            else:
                raise
        return None

    # ----------------------------------------------------------------------
    def __call__(self, r):
        if self.legacy == False:
            r.headers["X-Esri-Authorization"] = f"Bearer {self.token}"
        elif self.legacy and r.method == "GET":
            r.prepare_url(url=r.url, params={"token": self.token})
        elif self.legacy and r.method == "POST":
            data = parse_qs(r.body)
            data["token"] = self.token
            r.prepare_body(data, None, None)
        else:
            raise Exception(
                "Only POST and GET are supported with legacy methods of authentication."
            )
        r.register_hook("response", self.handle_401)
        try:
            self.pos = r.body.tell()
        except AttributeError:
            # In the case of HTTPKerberosAuth being reused and the body
            # of the previous request was a file-like object, pos has
            # the file position of the previous body. Ensure it's set to
            # None.
            self.pos = None
        # return request

        return r

    # ----------------------------------------------------------------------
    def handle_401(self, r, **kwargs):
        """
        handles the issues in the response where token might be rejected
        """
        parsed = parse_url(r.url)
        if (
            r.text.lower().find("invalid token") > -1
            or r.text.find("Token is valid but access is denied.") > -1
            or (parsed.scheme, parsed.netloc, parsed.path) in self._no_go_token
        ):
            # parsed = parse.urlparse(r.url)
            self._no_go_token.add((parsed.scheme, parsed.netloc, parsed.path))
            # Recreate the request without the token
            #
            r.content
            r.raw.release_conn()
            r.request.headers["referer"] = self._referer  # or "http"
            r.request.headers.pop("X-Esri-Authorization", None)
            _r = r.connection.send(r.request, **kwargs)
            _r.headers["referer"] = self._referer  # or "http"
            _r.headers.pop("X-Esri-Authorization", None)
            _r.history.append(r)
            return _r
        elif r.text.lower().find("token required") > -1:
            r.content
            r.raw.release_conn()
            r.request.headers["referer"] = self._referer  # or "http"
            r.headers["X-Esri-Authorization"] = f"Bearer {self.token}"
            _r = r.connection.send(r.request, **kwargs)
            _r.headers["referer"] = self._referer  # or "http"
            _r.headers["X-Esri-Authorization"] = f"Bearer {self.token}"
            _r.history.append(r)
            return _r
        return r

    # ----------------------------------------------------------------------
    def _refresh(self):
        """renews the token"""
        if self._response_type == "token":
            self._expiration_time = _dt.datetime.now() + _dt.timedelta(
                seconds=self._expiration - 1
            )
            self._init_response_type_token()
        else:
            self._auth_token = self._oauth.refresh_token(
                token_url=self._token_url,
                verify=self._verify_cert,
                client_id=self._oauth.client_id,
                expiration=20160,
            )
            if "expires_at" in self._auth_token:
                self._expiration_time = _dt.datetime.fromtimestamp(
                    self._auth_token["expires_at"]
                )
            elif "expires" in self._auth_token:
                self._expiration_time = _dt.datetime.now() + _dt.timedelta(
                    seconds=self._auth_token["expiration"]
                )
            else:
                self._expiration_time = _dt.datetime.now() + _dt.timedelta(seconds=300)


###########################################################################
class EsriGenTokenAuth(AuthBase, SupportMultiAuth):
    """
    This form of Authentication leverages the `generateToken` endpoint from
    the ArcGIS Product.  This is supported for ArcGIS Online, ArcGIS Enterprise
    and ArcGIS Server.

    This form of authentication is considered legacy and should only be used
    with unfederated server products.
    """

    _anon_urls = None
    _session = None
    _token = None
    _expires_on = None
    _portal_auth = None

    # ----------------------------------------------------------------------
    def __init__(
        self,
        token_url: str,
        referer: str,
        username: str = None,
        password: str = None,
        portal_auth: "EsriGenTokenAuth" = None,
        time_out: int = 1440,
        verify_cert: bool = True,
        legacy: bool = False,
        **kwargs,
    ) -> None:
        """init"""
        self.proxies = kwargs.pop("proxies", None)
        if username is None and portal_auth is None:
            raise Exception(
                "A portal_auth or username/password is required for GenerateToken"
            )
        self._anon_urls = set()
        self._legacy_auth = legacy
        self._portal_auth = portal_auth
        has_session = "session" in kwargs
        self._session = kwargs.pop("session", None) or requests.sessions.Session()
        self.verify_cert = verify_cert
        if has_session == False:
            self._session.verify = verify_cert
            self._session.allow_redirects = True
        if self.proxies:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                token_url = _token_url_validator(
                    _parse_arcgis_url(token_url),
                    session=self._session,
                    verify=False,
                    proxies=frozenset(self.proxies.items()),
                )
        else:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                token_url = _token_url_validator(
                    _parse_arcgis_url(token_url),
                    session=self._session,
                    verify=False,
                )
        self._thread_local = threading.local()

        self._expires_on = None
        self._token_url = token_url

        self.username = username
        self.password = password
        if referer is None and self.referer is None:
            self.referer = "http"
        else:
            self.referer = referer
        if not time_out or time_out < 5:
            self._time_out = 1440
        else:
            self._time_out = time_out

    # ----------------------------------------------------------------------
    def __str__(self):
        return f"<{self.__class__.__name__}, token=.....>"

    # ----------------------------------------------------------------------
    def __repr__(self):
        return f"<{self.__class__.__name__}, token=.....>"

    # ----------------------------------------------------------------------
    @property
    def referer(self) -> str:
        if "referer" in self._session.headers:
            return self._session.headers["referer"]
        else:
            self.referer = "http"
            return self.referer

    # ----------------------------------------------------------------------
    @referer.setter
    def referer(self, value: str) -> None:
        """Gets/Sets the referer"""
        if (
            "referer" in self._session.headers
            and self._session.headers["referer"] != value
        ):
            self._session.headers["referer"] = value
        else:
            self._session.headers["referer"] = value

    # ----------------------------------------------------------------------
    @property
    def time_out(self) -> int:
        """returns the time out time in minutes"""
        return self._time_out

    # ----------------------------------------------------------------------
    @time_out.setter
    def time_out(self, value: int):
        if value is None:
            value = 60
        if self._time_out != value:
            self._time_out = value

    # ----------------------------------------------------------------------
    @property
    def expiration(self) -> _dt.datetime:
        """Gets the time the token will expire on"""
        if self._expires_on:
            return self._expires_on
        else:
            self.token
            return self.expiration

    # ----------------------------------------------------------------------
    @cached(cache=TTLCache(maxsize=255, ttl=60))
    def token(self, server_url=None) -> str:
        if self._token:
            if (_dt.datetime.now() - _dt.timedelta(minutes=5)) >= self.expiration:
                self._token = None
            return self._token
        elif self.username and self.password:
            return self._init_token_auth_handshake()
        elif self._portal_auth:
            return self._init_token_auth_handshake(server_url)

        return None

    # ----------------------------------------------------------------------
    def __call__(self, r):
        """
        Handles the token authentication for the call
        """
        parsed = parse_url(r.url)
        server_url = f"{parsed.scheme}://{parsed.netloc}"
        self.init_per_thread_state()
        if self._legacy_auth and r.method == "GET":
            r.prepare_url(url=r.url, params={"token": self.token(server_url)})
        elif self._legacy_auth and r.method == "POST":
            data = parse_qs(r.body)
            data["token"] = self.token(server_url)
            r.prepare_body(data, None, None)
        else:
            r.headers["X-Esri-Authorization"] = f"Bearer {self.token(server_url)}"
        r.headers["referer"] = self.referer
        r.register_hook("response", self.handle_401)
        r.register_hook("response", self.handle_redirect)
        try:
            self.pos = r.body.tell()
        except AttributeError:
            # In the case of HTTPKerberosAuth being reused and the body
            # of the previous request was a file-like object, pos has
            # the file position of the previous body. Ensure it's set to
            # None.
            self.pos = None
        # return request
        self._thread_local.num_401_calls = 1
        return r

    # ----------------------------------------------------------------------
    def init_per_thread_state(self):
        # Ensure state is initialized just once per-thread
        if not hasattr(self._thread_local, "init"):
            self._thread_local.init = True
            self._thread_local.last_nonce = ""
            self._thread_local.nonce_count = 0
            self._thread_local.chal = {}
            self._thread_local.pos = None
            self._thread_local.num_401_calls = None

    # ----------------------------------------------------------------------

    def handle_401(self, r, **kwargs):
        # if r.status_code in [401, 402, 403]:
        # raise Exception(f"Error: {r.status_code}, {r.text}")
        if not 400 <= r.status_code < 500 and (
            r.status_code == 200 and r.text.find("Invalid Token") == -1
        ):
            self._thread_local.num_401_calls = 1
            return r
        elif r.status_code == 200 and r.text.find("Invalid Token") > -1:
            if self._thread_local.num_401_calls < 2:
                self._thread_local.num_401_calls += 1
                # Consume content and release the original connection
                # to allow our new request to reuse the same one.
                r.content
                r.close()
                prep = r.request.copy()
                if self._legacy_auth and prep.method == "GET":
                    parsed = parse_url(prep.url)
                    url = urlunparse(
                        (
                            parsed.scheme,
                            parsed.netloc,
                            parsed.path,
                            "",
                            "",
                            "",
                        )
                    )
                    kv = dict(parse_qsl(parsed.query))
                    kv.pop("token", None)
                    prep.prepare_url(url=url, params=kv)
                elif self._legacy_auth and prep.method == "POST":
                    data = parse_qs(prep.body)
                    data.pop("token", None)
                    prep.prepare_body(data, None, None)

                extract_cookies_to_jar(prep._cookies, r.request, r.raw)
                prep.prepare_cookies(prep._cookies)
                prep.headers.pop("X-Esri-Authorization", None)
                _r = r.connection.send(prep, **kwargs)
                _r.history.append(r)
                _r.request = prep
                return _r
        return r

    # ----------------------------------------------------------------------

    def handle_redirect(self, r, **kwargs):
        if r.is_redirect:
            self._thread_local.num_401_calls = 1

    # ----------------------------------------------------------------------
    @cached(cache=TTLCache(maxsize=255, ttl=60))
    def _init_token_auth_handshake(self, server_url=None):
        """gets the token"""
        if self.username and self.password:  # Basic Generate Token Logic
            self.time_out = 60
            postdata = {
                "username": self.username,
                "password": self.password,
                "referer": self.referer,
                "client": "referer",
                "expiration": 60,  # self.time_out,
                "f": "json",
            }
            resp = self._session.post(
                url=self._token_url,
                data=postdata,
                verify=self.verify_cert,
                proxies=self.proxies,
            )
            if resp.status_code == 200:
                data = resp.json()
                if "error" in data:
                    raise Exception(data)
                if "expires" in data:
                    self._expires_on = _dt.datetime.fromtimestamp(
                        int(data["expires"]) / 1000
                    )
                elif "expiration" in data:
                    self._expires_on = _dt.datetime.fromtimestamp(
                        int(data["expiration"])
                    )
                else:
                    self._expires_on = _dt.datetime.now() + _dt.timedelta(
                        minutes=self._time_out - 3
                    )
                self._token = data["token"]
                return self._token
            else:
                raise Exception("Could not generate the token")
        elif self._portal_auth:  # Federated Token Logic
            postdata = {
                "serverURL": server_url,
                "token": self._portal_auth.token(),
                "expiration": str(self.time_out),
                "f": "json",
                "request": "getToken",
                "referer": self.referer,
            }
            resp = self._session.post(
                url=self._token_url,
                data=postdata,
                verify=self.verify_cert,
                proxies=self.proxies,
            )
            if resp.status_code == 200:
                data = resp.json()
                if "error" in data:
                    raise Exception(data)
                if "expires" in data:
                    self._expires_on = _dt.datetime.fromtimestamp(
                        int(data["expires"]) / 1000
                    )
                elif "expiration" in data:
                    self._expires_on = _dt.datetime.fromtimestamp(
                        int(data["expiration"])
                    )
                else:
                    self._expires_on = _dt.datetime.now() + _dt.timedelta(
                        minutes=self._time_out - 3
                    )
                self._token = data["token"]
                return self._token
        else:
            raise Exception("Invalid Credentials")



from typing import Union, Optional, ClassVar
from dataclasses import field, dataclass

from arcgis.realtime import Velocity
from arcgis.realtime.velocity.feeds._feed_template import _FeedTemplate
from arcgis.realtime.velocity.feeds.geometry import (
    _HasGeometry,
    SingleFieldGeometry,
    XYZGeometry,
)
from arcgis.realtime.velocity.feeds.time import _HasTime, TimeInstant, TimeInterval
from arcgis.realtime.velocity.input.format import (
    DelimitedFormat,
    _format_from_config,
)


@dataclass
class HttpSimulator(_FeedTemplate, _HasTime, _HasGeometry):
    """
    Simulate events from a text file. This data class can be used to define the feed configuration and to create
    the feed.

    ============================    ====================================================================
    **Parameter**                    **Description**
    ----------------------------    --------------------------------------------------------------------
    label                           String. Unique label for this feed instance.
    ----------------------------    --------------------------------------------------------------------
    description                     String. Feed description.
    ----------------------------    --------------------------------------------------------------------
    url                             String. The full URL to the externally accessible simulation file.
    ----------------------------    --------------------------------------------------------------------
    field_separator                 String. The character, or delimiter, which separates field values in
                                    the simulation file. The default is: ,.
    ----------------------------    --------------------------------------------------------------------
    features_per_execution          int. The number of records (features) to simulate at a time. The default
                                    is: 1.
    ----------------------------    --------------------------------------------------------------------
    interval_for_sending_events     int. The interval between sending the number of features per execution.
                                    The default is: 1000.
    ----------------------------    --------------------------------------------------------------------
    repeat_simulation               boolean. Whether to automatically restart from the beginning when
                                    the end of the file is reached. The default is: True.
    ----------------------------    --------------------------------------------------------------------
    time_field_index                int. The numerical index of the date field in the dataset, where the
                                    index starts at 0. The default is: 0.
    ----------------------------    --------------------------------------------------------------------
    convert_to_current_time         boolean. Whether to convert the time values in the dataset to
                                    current time as the data is simulated. The default is: True.
    ============================    ====================================================================

    ============================    ====================================================================
    **Optional Argument**           **Description**
    ============================    ====================================================================
    data_format                     [:class:`~arcgis.realtime.velocity.input.DelimitedFormat`].
                                    An instance that contains the data format
                                    configuration for this feed. Configure only allowed formats.
                                    If this is not set right during initialization, a format will be
                                    auto-detected and set from a sample of the incoming data. This sample
                                    will be fetched from the configuration provided so far in the init.
    ----------------------------    --------------------------------------------------------------------
    track_id_field                  String. Name of the field from the incoming data that should be set as
                                    track ID.
    ----------------------------    --------------------------------------------------------------------
    geometry                        [:class:`~arcgis.realtime.velocity.feeds.XYZGeometry`, :class:`~arcgis.realtime.velocity.feeds.SingleFieldGeometry`]. An instance of geometry configuration
                                    that will be used to create geometry objects from the incoming data.
    ----------------------------    --------------------------------------------------------------------
    time                            [:class:`~arcgis.realtime.velocity.feeds.TimeInstant`, :class:`~arcgis.realtime.velocity.feeds.TimeInterval`]. An instance of time configuration that
                                    will be used to create time information from the incoming data.
    ============================    ====================================================================

    :return: A data class with Http simulator feed configuration.

    .. code-block:: python

        # Usage Example

        from arcgis.realtime.velocity.feeds import HttpSimulator
        from arcgis.realtime.velocity.input.format import DelimitedFormat
        from arcgis.realtime.velocity.feeds.geometry import XYZGeometry, SingleFieldGeometry
        from arcgis.realtime.velocity.feeds.time import TimeInterval, TimeInstant

        http_simulator = HttpSimulator(
            label="feed_name",
            description="feed_description",
            url="http_simulator_url",
            data_format=None
        )

        # use velocity object to get the FeedsManager instance
        feeds = velocity.feeds

        # use the FeedsManager object to create a feed from this feed configuration
        http_simulator_feed = feeds.create(http_simulator)
        http_simulator_feed.start()
        feeds.items

    """

    # fields that the user sets during init
    # HTTP Simulator specific properties
    url: str
    field_separator: str = field(default=",")
    features_per_execution: int = field(default=1)
    interval_for_sending_events: int = field(default=1000)
    repeat_simulation: bool = field(default=True)
    time_field_index: int = field(default=0)
    convert_to_current_time: bool = field(default=True)

    # user can define these properties even after initialization
    data_format: Optional[Union[DelimitedFormat]] = None
    # FeedTemplate properties
    track_id_field: Optional[str] = None
    # HasGeometry properties
    geometry: Optional[Union[XYZGeometry, SingleFieldGeometry]] = None
    # HasTime properties
    time: Optional[Union[TimeInstant, TimeInterval]] = None

    # FeedTemplate properties
    _name: ClassVar[str] = "simulator"

    def __post_init__(self):
        if Velocity is None:
            return
        self._util = Velocity._util

        # validation of fields
        if self._util.is_valid(self.label) == False:
            raise ValueError(
                "Label should only contain alpha numeric, _ and space only"
            )

        # generate dictionary of this feed object's properties that will be used to query test-connection and
        # sample-messages Rest endpoint
        feed_properties = self._generate_feed_properties()

        test_connection = self._util.test_connection(
            input_type="feed", payload=feed_properties
        )
        # Test connection to make sure feed can fetch schema
        if test_connection is True:
            # test connection succeeded. Now try getting sample messages
            sample_payload = {
                "properties": {
                    "maxSamplesToCollect": 5,
                    "timeoutInMillis": 5000,
                }
            }
            self._dict_deep_merge(sample_payload, feed_properties)
            # Sample messages to fetch schema/fields
            sample_messages = self._util.sample_messages(
                input_type="feed", payload=sample_payload
            )

            if sample_messages["featureSchema"] is not None:
                # sample messages succeeded. Get Feature Schema from it
                self._set_fields(sample_messages["featureSchema"])

            # if Format was not specified by user, use the auto-detected format from the sample messages response as this feed object's format.
            if self.data_format is None:
                self.data_format = _format_from_config(sample_messages)

            print(
                "Feature Schema retrieved from the Feed:",
                sample_messages["featureSchema"],
            )

            # initiate actions for each of the following properties if it was set at init
            if self.track_id_field is not None:
                self.set_track_id(self.track_id_field)
            if self.geometry is not None:
                self.set_geometry_config(self.geometry)
            if self.time is not None:
                self.set_time_config(self.time)

    def _build(self) -> dict:
        feed_configuration = {
            "id": "",
            "label": self.label,
            "description": self.description,
            "feed": {**self._generate_schema_transformation()},
            "properties": {"executable": True},
        }

        feed_properties = self._generate_feed_properties()
        self._dict_deep_merge(feed_configuration["feed"], feed_properties)
        print(feed_configuration)
        return feed_configuration

    def _generate_feed_properties(self) -> dict:
        feed_properties = {
            "name": self._name,
            "properties": {
                f"{self._name}.url": self.url,
                f"{self._name}.fieldSeparator": self.field_separator,
                f"{self._name}.featuresPerTime": self.features_per_execution,
                f"{self._name}.repeat": self.repeat_simulation,
                f"{self._name}.intervalInMillis": self.interval_for_sending_events,
                f"{self._name}.timeFieldIndex": self.time_field_index,
                f"{self._name}.setToCurrentTime": self.convert_to_current_time,
            },
        }

        if self.data_format is not None:
            format_dict = self.data_format._build()
            self._dict_deep_merge(feed_properties, format_dict)

        return feed_properties



import torch
import torch.nn as nn


class _TSLSTM(nn.Module):
    def __init__(
        self,
        input_size=1,
        output_size=1,
        device=None,
        hidden_layer_size=100,
        num_layers=1,
        dropout=0.2,
        bidirectional=False,
    ):
        super(_TSLSTM, self).__init__()
        self.device = device
        self.input = input_size
        self.hidden = hidden_layer_size
        self.num_layers = num_layers
        self.multiplier = 1
        if bidirectional:
            self.multiplier = 2
        if self.num_layers > 1:
            self.lstm = nn.LSTM(
                input_size=self.input,
                hidden_size=self.hidden,
                num_layers=self.num_layers,
                batch_first=True,
                bidirectional=bidirectional,
                dropout=dropout,
            )
        else:
            self.lstm = nn.LSTM(
                input_size=self.input,
                hidden_size=self.hidden,
                num_layers=self.num_layers,
                batch_first=True,
                bidirectional=bidirectional,
            )
        self.dropout = nn.Dropout(dropout)
        self.linear = nn.Linear(self.multiplier * self.hidden, output_size)

    def forward(self, input_seq):
        input_seq = input_seq.permute(0, 2, 1)
        batch_size = input_seq.size(0)
        x = input_seq.to(torch.float)
        h0 = torch.zeros(self.multiplier * self.num_layers, batch_size, self.hidden).to(
            self.device
        )
        c0 = torch.zeros(self.multiplier * self.num_layers, batch_size, self.hidden).to(
            self.device
        )
        out, (h0, c0) = self.lstm(x, (h0, c0))
        if self.multiplier == 1:
            h_comb = h0[-1]
        else:
            h_comb = torch.cat([h0[: self.num_layers], h0[self.num_layers :]], dim=-1)[
                -1
            ]
        outputs = self.linear(h_comb)
        return outputs.to(self.device)



import os
import json
import random
import logging
import tempfile
import datetime
import traceback
from pathlib import Path
from .text_data import TextDataObject

try:
    import spacy
    from spacy.gold import offsets_from_biluo_tags as _offsets_from_biluo_tags
    from spacy.gold import iob_to_biluo as _iob_to_biluo
    import pandas as pd
    import numpy as np

    HAS_SPACY = True
except Exception as e:
    spacy_exception = "\n".join(traceback.format_exception(type(e), e, e.__traceback__))
    HAS_SPACY = False


__all__ = ["_from_iob_tags", "_from_json", "_NERData", "even_mults", "_timelapsed"]


def _raise_spacy_import_error():
    error_message = (
        f"{spacy_exception}\n\n\n"
        "This module requires spacy version 2.1.8 or above and fastprogress."
        "Install it using 'pip install spacy==2.1.8 fastprogress pandas'"
    )
    raise Exception(error_message)


def even_mults(
    start: float, stop: float, n: int
):  # Taken from FastAI(https://github.com/fastai/fastai/blob/master/fastai/core.py#L150)
    "Build log-stepped array from `start` to `stop` in `n` steps."
    mult = stop / start
    step = mult ** (1 / (n - 1))
    return np.array([start * (step**i) for i in range(n)])


def _timelapsed(t_start):
    """returns timedelta in hh:mm:ss format"""
    b = datetime.datetime.now() - t_start
    h, r = divmod(b.seconds, 3600)
    m, s = divmod(r, 60)
    return "%02d:%02d:%02d" % (h, m, s)


def _from_iob_tags(tokens_collection, tags_collection):
    """
    Converts training data from ``IOB`` format to spacy offsets.

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    tokens_collection       Required [list]. List of token lists
                            Example: [[This,is,a,test],[This,is,a,test1]]
    ---------------------   -------------------------------------------
    tags_collection         Required [list]. List of tag lists
                            Example: [[B-tagname,O,O,O],[B-tagname,I-tagname,O,O]]
    =====================   ===========================================
    """

    train_data = []
    nlp = spacy.blank("en")
    for tags, tokens in zip(tags_collection, tokens_collection):
        try:
            tags = _iob_to_biluo(tags)

            doc = spacy.tokens.doc.Doc(
                nlp.vocab, words=tokens, spaces=[True] * (len(tokens) - 1) + [False]
            )
            # run the standard pipeline against it
            for name, proc in nlp.pipeline:
                doc = proc(doc)

            text = " ".join(tokens)
            tags = _offsets_from_biluo_tags(doc, tags)
            train_data.append((text, {"entities": tags}))
        except:
            pass

    return train_data


def _from_json(path, text_key="text", offset_key="labels", encoding="UTF-8"):
    """
    Converts training data from JSON format to spacy offsets.

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    text_key                Optional:str='text. Json key under which text is available
    ---------------------   -------------------------------------------
    offset_key              Optional:str='labels. Json key under which offsets are available
    =====================   ===========================================
    json-schema:
    ----------
    {"id": 1, "text": "EU rejects ...", "labels": [[0,2,"ORG"], [11,17, "MISC"], [34,41,"ORG"]]}
    {"id": 2, "text": "Peter Blackburn", "labels": [[0, 15, "PERSON"]]}
    {"id": 3, "text": "President Obama", "labels": [[10, 15, "PERSON"]]}
    ----------
    returns: A list that can be consumed by ner_databunch.
    """

    train_data = []
    with open(path, "r", encoding=encoding) as f:
        data_list = f.readlines()
    for i, item in enumerate(data_list):
        try:
            execute_second_exeption = False
            train_data.append(
                (json.loads(item)[text_key], {"entities": json.loads(item)[offset_key]})
            )
        except KeyError as key:
            _key = key
            execute_second_exeption = True
        if execute_second_exeption:
            try:
                train_data.append(
                    (json.loads(item)["data"], {"entities": json.loads(item)["label"]})
                )
            except KeyError as key:
                raise Exception(
                    f"{_key} key not present in record {i} of input json file."
                )

    return train_data


def _get_tags_and_tokens_collection(path, ignore_tag_order=False, encoding="UTF-8"):
    unique_tags = set()
    tags_collection, tokens_collection = [], []
    tags_df = pd.read_csv(path / "tags.csv", encoding=encoding, dtype="str")
    tokens_df = pd.read_csv(path / "tokens.csv", encoding=encoding, dtype="str")

    for i, tags in tags_df.iterrows():
        if ignore_tag_order:
            tags = [x.split("-")[-1] for x in tags.dropna()]
        else:
            tags = [x for x in tags.dropna()]
        unique_tags.update(tags)
        tags_collection.append(tags)

    for i, tokens in tokens_df.iterrows():
        tokens_collection.append(list(tokens.dropna()))

    return tags_collection, tokens_collection, unique_tags


class _NERData:
    working_dir = None
    """
    #     Prepares a data object
    #
    #     =====================   ===========================================
    #     **Parameter**            **Description**
    #     ---------------------   -------------------------------------------
    #     dataset_type            Required string. ['ner_json', 'IOB', 'BILUO']
    #     ---------------------   -------------------------------------------
    #     address_tag             Optional dict. Address field/tag name
    #                             in the training data.
    #     ---------------------   -------------------------------------------
    #     val_split_pct           Optional Float. Percentage of training data to keep
    #                             as validation. The default value is 0.1.
    #     =====================   ===========================================
    #     returns: A list [text,{entities},text,{entities}] that can be ingested by ``EntityRecognizer``.
    """

    def __init__(
        self,
        dataset_type,
        path,
        batch_size,
        class_mapping=None,
        seed=42,
        val_split_pct=0.1,
        encoding="UTF-8",
    ):
        self.dataset_type = dataset_type
        self.path = path
        self.data = None
        self.backbone = None
        self.batch_size = batch_size
        self.class_mapping = class_mapping
        self.seed = seed
        self.val_split_pct = val_split_pct
        self.encoding = encoding
        self.prepare_data_for_spacy()

    def show_batch(self):
        return self.data.show_batch()

    def get_data_object(self):
        data = self.data
        if self.working_dir is None:
            path = getattr(data, "path", self.path)
            if os.path.isfile(path):
                path = Path(os.path.dirname(os.path.abspath(path)))
            data.working_dir = path
            from .._data import _prepare_working_dir

            _prepare_working_dir(path)
        else:
            data.working_dir = self.working_dir

        return data

    def prepare_data_for_transformer(self, ignore_tag_order=True, label2id=None):
        unique_tags = set()
        path = Path(self.path)
        tags_collection, tokens_collection = [], []
        if self.class_mapping:
            address_tag = self.class_mapping.get("address_tag")
        else:
            address_tag = "Address"

        if self.dataset_type == "ner_json":
            # converts json schema to a list of list of tokens and tags
            # json - schema:
            # ----------
            # {"id": 1, "text": "Officers were dispatched ...", "labels": [[30, 38, "Crime"], [45, 92, "Address"], ...]}
            # ----------
            # converted form:
            # [('Officers were dispatched to a', 'O'),
            #  ('robbery', 'Crime'),
            #  ('of the', 'O'),
            #  ('Associated Bank in the 1500 block of W Broadway', 'Address')]
            data_list = _from_json(path=path, encoding=self.encoding)
            for i, row in enumerate(data_list):
                prev_start = 0
                tmp_tags_list, tmp_tokens_list = [], []
                text, labels = row[0], row[1]["entities"]
                for item in sorted(labels, key=lambda x: x[0]):
                    c_text = text[prev_start : item[0]].strip()
                    tmp_tags_list.append("O")
                    unique_tags.add("O")
                    tmp_tokens_list.append(c_text)
                    c_text = text[item[0] : item[1]].strip()
                    tmp_tags_list.append(item[2])
                    tmp_tokens_list.append(c_text)
                    unique_tags.add(item[2])
                    prev_start = item[1]

                tags_collection.append(tmp_tags_list)
                tokens_collection.append(tmp_tokens_list)

        elif self.dataset_type in ["BIO", "IOB", "LBIOU", "BILUO"]:
            (
                tags_collection,
                tokens_collection,
                unique_tags,
            ) = _get_tags_and_tokens_collection(
                path, ignore_tag_order=ignore_tag_order, encoding=self.encoding
            )

            if ignore_tag_order:
                unique_tags = set({x.split("-")[-1] for x in unique_tags})
        else:
            error_message = (
                f"Wrong argument - {self.dataset_type} supplied for `dataset_type` parameter. "
                "Valid values are - 'ner_json', 'BIO', 'IOB', 'LBIOU' and 'BILUO'"
            )
            raise Exception(error_message)

        # unique_tags are formed after reading the data, label2id is a mapping of tags to numbers
        # provided by the model. If they differ then we have to initialize the model head according
        # to the tags provided in the data
        if label2id and unique_tags != label2id.keys():
            # label2id = None
            error_message = (
                f"Looks like the backbone is fine-tuned on the following entities - {list(label2id.keys())}"
                f" and your data consists of the following entities - `{unique_tags}`."
                f"\nPlease use a base model of the backbone and fine-tune it on your data or use a "
                f"model which is fine-tuned on a data having same labels as - `{unique_tags}`"
            )
            raise Exception(error_message)
        self.data = TextDataObject.prepare_data_for_entity_recognition(
            tokens_collection=tokens_collection,
            tags_collection=tags_collection,
            address_tag=address_tag,
            unique_tags=unique_tags,
            seed=self.seed,
            batch_size=self.batch_size,
            val_split_pct=self.val_split_pct,
            label2id=label2id,
        )
        self.backbone = "transformers"

    def prepare_data_for_spacy(self):
        if not HAS_SPACY:
            _raise_spacy_import_error()

        random.seed(self.seed)
        v_list = spacy.__version__.split(".")
        version = sum([int(j) * 10 ** (2 * i) for i, j in enumerate(v_list[::-1])])
        if version < 20108:  # checking spacy version
            error_message = (
                "Entity recognition model needs spacy version 2.1.8 or higher."
                f"Your current spacy version is {spacy.__version__}, please update using 'pip install'"
            )
            return logging.error(error_message)

        path = Path(self.path)
        train_data = []

        if self.class_mapping:
            address_tag = self.class_mapping.get("address_tag")
        else:
            address_tag = "Address"

        if self.dataset_type == "ner_json":
            train_data = _from_json(path=path, encoding=self.encoding)
            path = path.parent
        elif self.dataset_type == "BIO" or self.dataset_type == "IOB":
            tags_collection, tokens_collection, _ = _get_tags_and_tokens_collection(
                path, encoding=self.encoding
            )
            train_data = _from_iob_tags(
                tags_collection=tags_collection, tokens_collection=tokens_collection
            )
        elif self.dataset_type == "LBIOU" or self.dataset_type == "BILUO":
            nlp = spacy.blank("en")
            tags_collection, tokens_collection, _ = _get_tags_and_tokens_collection(
                path, encoding=self.encoding
            )

            for tags, tokens in zip(tags_collection, tokens_collection):
                try:
                    tags = _iob_to_biluo(tags)

                    doc = spacy.tokens.doc.Doc(
                        nlp.vocab,
                        words=tokens,
                        spaces=[True] * (len(tokens) - 1) + [False],
                    )
                    # run the standard pipeline against it
                    for name, proc in nlp.pipeline:
                        doc = proc(doc)
                    text = " ".join(tokens)
                    tags = _offsets_from_biluo_tags(doc, tags)
                    train_data.append((text, {"entities": tags}))
                except Exception as exception:
                    raise Exception(f"Exception while preparing data : {exception} ")
                    pass
        else:
            error_message = (
                f"Wrong argument - {self.dataset_type} supplied for `dataset_type` parameter. "
                "Valid values are - 'ner_json', 'BIO', 'IOB', 'LBIOU' and 'BILUO'"
            )
            raise Exception(error_message)

        self.data = spaCyNERDatabunch(
            train_data,
            val_split_pct=self.val_split_pct,
            batch_size=self.batch_size,
            address_tag=address_tag,
            test_ds=None,
        )
        self.data.path = path
        self.backbone = "spacy"


class _spaCyNERItemlist:
    """
    Creates a dataset to store data within ``ner_databunch`` object.

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    batch_size              Batch size.
    ---------------------   -------------------------------------------
    data                    Required: list of tuple containing text and its entities.
    =====================   ===========================================

    :return: dataset.
    """

    def __init__(self, batch_size, data):
        self.batch_size = batch_size
        self.entities = list(
            {
                i[2]
                for i in pd.concat(
                    [pd.Series(i["entities"]) for i in [o[1] for o in data]]
                )
            }
        )  ##Extracting all the unique entity names from input json
        self.data = data
        self.x = [o[0] for o in data]
        self.y = [o[1] for o in data]

    def __getitem__(self, i):
        return self.data[i]

    def __len__(self):
        return len(self.data)

    def _random_batch(self, data):
        res = []
        for j in range(self.batch_size):
            res.append(random.choice(data))
        return res

    def _entities_to_dataframe(self, item):
        """
        This function is used to create pandas dataframe from training input data json.
        """
        text = item[0]
        df = pd.DataFrame(item[1].get("entities"))

        out_dict = {}
        if len(df):
            for x in df[2].unique():
                out_dict[x] = df[df[2] == x][[0, 1]].values.tolist()

        out = {}
        out["text"] = text
        for key in out_dict.keys():
            for tpl in out_dict.get(key):
                if out.get(key) == None:
                    out[key] = []
                out[key].append(text[tpl[0] : tpl[1]])
        return pd.Series(out)

    def show_batch(self):
        """
        This function shows a batch from the _spaCyNERItemlist.
        """
        data = self._random_batch(self.data)
        lst = []
        for item in data:
            lst.append(self._entities_to_dataframe(item))
        batch_df = pd.concat(lst, axis=1, sort=True).T

        text = batch_df["text"]
        batch_df.drop("text", axis=1, inplace=True)
        batch_df.insert(loc=0, column="text", value=text)
        return batch_df.fillna("")


class spaCyNERDatabunch:

    """
    Creates a databunch object.

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    ds                      Required data list returned from _ner_prepare_data().
    ---------------------   -------------------------------------------
    val_split_pct           Optional float. Percentage of training data to keep
                            as validation.
                            The default Value is 0.1.
    ---------------------   -------------------------------------------
    batch_size              Optional integer. Batch size
                            The default value is 5.
    =====================   ===========================================

    :return: dataset
    """

    def __init__(self, ds, val_split_pct, batch_size, test_ds=None, address_tag=None):
        random.shuffle(ds)
        # creating an _spaCyNERItemlist with training dataset
        self.train_ds = _spaCyNERItemlist(
            batch_size, data=ds[: int(len(ds) * (1 - val_split_pct))]
        )
        # creating an _spaCyNERItemlist with validation dataset
        self.val_ds = _spaCyNERItemlist(
            batch_size, data=ds[int(len(ds) * (1 - val_split_pct)) :]
        )
        self.entities = list(
            set(self.train_ds.entities).union(set(self.val_ds.entities))
        )
        self._address_tag = address_tag
        self._has_address = True
        self.batch_size = batch_size
        if self.batch_size > len(self.train_ds):
            error_message = (
                f"Number of training data items ({len(self.train_ds)}) "
                f"is less than the batch size ({self.batch_size}). "
                "Please get more training data or lower the batch size"
            )
            logging.error(error_message)
        if self._address_tag not in self.entities:
            self._has_address = False
            logging.warning(
                "No Address tag found in your data.\n\
                1. If your data has an address field, pass your address field name as address tag in class mapping \n\
                e.g. - data=prepare_data(dataset_type=ds_type,path=training_data_folder,\n\t\t\t\
                    class_mapping={address_tag:address_field_name})\n\
                2. Else no action is required, if your data does not have any address information."
            )

    def show_batch(self):
        return self.train_ds.show_batch()



"""
The ``Functions`` module is used to take :class:`~arcgis.geometry.Geometry` types as parameters and return
:class:`~arcgis.geometry.Geometry` type results.
"""
from __future__ import annotations
from enum import Enum
import json
from typing import Any, Optional, Union
from arcgis.geometry import (
    Geometry,
    Point,
    MultiPoint,
    Polyline,
    Polygon,
    SpatialReference,
)
import arcgis.env
from arcgis.gis import GIS


class AreaUnits(Enum):
    """
    Represents the Supported Geometry Service Area Units Enumerations.
    Example: areas_and_lengths(polygons=[geom],area_unit=AreaUnits.ACRES)
    """

    UNKNOWNAREAUNITS = {"areaUnit": "esriUnknownAreaUnits"}
    SQUAREINCHES = {"areaUnit": "esriSquareInches"}
    SQUAREFEET = {"areaUnit": "esriSquareFeet"}
    SQUAREYARDS = {"areaUnit": "esriSquareYards"}
    ACRES = {"areaUnit": "esriAcres"}
    SQUAREMILES = {"areaUnit": "esriSquareMiles"}
    SQUAREMILLIMETERS = {"areaUnit": "esriSquareMillimeters"}
    SQUARECENTIMETERS = {"areaUnit": "esriSquareCentimeters"}
    SQUAREDECIMETERS = {"areaUnit": "esriSquareDecimeters"}
    SQUAREMETERS = {"areaUnit": "esriSquareMeters"}
    ARES = {"areaUnit": "esriAres"}
    HECTARES = {"areaUnit": "esriHectares"}
    SQUAREKILOMETERS = {"areaUnit": "esriSquareKilometers"}


class LengthUnits(Enum):
    """
    Represents the Geometry Service Length Units Enumerations
    Example: areas_and_lengths(polygons=[geom],length_unit=LengthUnits.FOOT)
    """

    BRITISH1936FOOT = 9095
    GOLDCOASTFOOT = 9094
    INTERNATIONALCHAIN = 9097
    INTERNATIONALLINK = 9098
    INTERNATIONALYARD = 9096
    STATUTEMILE = 9093
    SURVEYYARD = 109002
    FIFTYKMLENGTH = 109030
    ONEFIFTYKMLENGTH = 109031
    DECIMETER = 109005
    CENTIMETER = 1033
    MILLIMETER = 1025
    INTERNATIONALINCH = 109008
    USSURVEYINCH = 109009
    INTERNATIONALROD = 109010
    USSURVEYROD = 109011
    USNAUTICALMILE = 109012
    UKNAUTICALMILE = 109013
    METER = 9001
    GERMANMETER = 9031
    FOOT = 9002
    SURVEYFOOT = 9003
    CLARKEFOOT = 9005
    FATHOM = 9014
    NAUTICALMILE = 9030
    SURVEYCHAIN = 9033
    SURVEYLINK = 9034
    SURVEYMILE = 9035
    KILOMETER = 9036
    CLARKEYARD = 9037
    CLARKECHAIN = 9038
    CLARKELINK = 9039
    SEARSYARD = 9040
    SEARSFOOT = 9041
    SEARSCHAIN = 9042
    SEARSLINK = 9043
    BENOIT1895A_YARD = 9050
    BENOIT1895A_FOOT = 9051
    BENOIT1895A_CHAIN = 9052
    BENOIT1895A_LINK = 9053
    BENOIT1895B_YARD = 9060
    BENOIT1895B_FOOT = 9061
    BENOIT1895B_CHAIN = 9062
    BENOIT1895B_LINK = 9063
    INDIANFOOT = 9080
    INDIAN1937FOOT = 9081
    INDIAN1962FOOT = 9082
    INDIAN1975FOOT = 9083
    INDIANYARD = 9084
    INDIAN1937YARD = 9085
    INDIAN1962YARD = 9086
    INDIAN1975YARD = 9087
    FOOT1865 = 9070
    RADIAN = 9101
    DEGREE = 9102
    ARCMINUTE = 9103
    ARCSECOND = 9104
    GRAD = 9105
    GON = 9106
    MICRORADIAN = 9109
    ARCMINUTECENTESIMAL = 9112
    ARCSECONDCENTESIMAL = 9113
    MIL6400 = 9114


# -------------------------------------------------------------------------
def areas_and_lengths(
    polygons: Polygon,
    length_unit: str | LengthUnits,
    area_unit: str | AreaUnits,
    calculation_type: str,
    spatial_ref: int = 4326,
    gis: Optional[GIS] = None,
    future: bool = False,
):
    """
    The ``areas_and_lengths`` function calculates areas and perimeter lengths
    for each :class:`~arcgis.geometry.Polygon` specified in the input array.

    ================  ===============================================================================
    **Keys**          **Description**
    ----------------  -------------------------------------------------------------------------------
    polygons          The array of :class:`~arcgis.geometry.Polygon` whose areas and lengths are to be computed.
    ----------------  -------------------------------------------------------------------------------
    length_unit       The length unit in which the perimeters of
                      polygons will be calculated. If ``calculation_type``
                      is planar, then ``length_unit`` can be any esriUnits
                      constant (string or integer). If ``calculationType`` is
                      not planar, then ``length_unit`` must be a linear
                      esriUnits constant, such as `esriSRUnit_Meter`(i.e. `9001`|`LengthUnits.METER`) or
                      `esriSRUnit_SurveyMile`(i.e. `9035`|`LengthUnits.SURVEYMILE`). If ``length_unit`` is not
                      specified, the units are derived from ``spatial_ref``. If ``spatial_ref`` is not
                      specified as well, the units are in meters. For a list of
                      valid units, see `esriSRUnitType Constants` and
                      `esriSRUnit2Type Constants`.
    ----------------  -------------------------------------------------------------------------------
    area_unit         The area unit in which areas of polygons will be
                      calculated. If calculation_type is planar, then
                      area_unit can be any `esriAreaUnits` constant (dict or enum). If ``calculation_type`` is
                      not planar, then ``area_unit`` must be a `esriAreaUnits` constant such
                      as `AreaUnits.SQUAREMETERS` (i.e. `{"areaUnit": "esriSquareMeters"}`) or
                      `AreaUnits.SQUAREMILES` (i.e. `{"areaUnit": "esriSquareMiles"}`). If
                      ``area_unit`` is not specified, the units are derived
                      from ``spatial_ref``. If ``spatial_ref`` is not specified, then the units are in square
                      meters. For a list of valid units, see
                      `esriAreaUnits Constants`.
                      The list of valid esriAreaUnits constants include,
                      `esriSquareInches | esriSquareFeet |
                      esriSquareYards | esriAcres | esriSquareMiles |
                      esriSquareMillimeters | esriSquareCentimeters |
                      esriSquareDecimeters | esriSquareMeters | esriAres
                      | esriHectares | esriSquareKilometers.`
    ----------------  -------------------------------------------------------------------------------
    calculation_type  The type defined for the area and length calculation of the input geometries. The type can be one
                      of the following values:

                          1. planar - Planar measurements use 2D Euclidean distance to calculate area and length. This
                          should only be used if the area or length needs to be calculated in the given
                          :class:`~arcgis.geometry.SpatialReference`. Otherwise, use ``preserveShape``.

                          2. geodesic - Use this type if you want to calculate an area or length using only the vertices
                          of the :class:`~arcgis.geometry.Polygon` and define the lines between the points as geodesic
                          segments independent of the actual shape of the :class:`~arcgis.geometry.Polygon`. A geodesic
                          segment is the shortest path between two points on an ellipsoid.

                          3. preserveShape - This type calculates the area or length of the geometry on the surface of
                          the Earth ellipsoid. The shape of the geometry in its coordinate system is preserved.
    ----------------  -------------------------------------------------------------------------------
     future           Optional boolean. If True, a future object will be returned and the process
                      will not wait for the task to complete. The default is False, which means wait for results.
    ================  ===============================================================================

    :returns:
        A JSON as dictionary, or a `GeometryJob` object. If ``future = True``,
        then the result is a :class:`~concurrent.futures.Future` object. Call ``result()`` to get the response.

    .. code-block:: python

            >>> # Use case 1
            >>> areas_and_lengths(polygons =[polygon1, polygon2,...],
                                  length_unit = 9001,
                                  area_unit = {"areaUnit": "esriSquareMeters"},
                                  calculation_type = "planar")
            >>> # Use case 2
            >>> from arcgis.geometry import LengthUnits, AreaUnits
            >>> areas_and_lengths(polygons =[polygon1, polygon2,...],
                                  length_unit = LengthUnits.METER,
                                  area_unit = AreaUnits.SQUAREMETERS,
                                  calculation_type = "planar",
                                  future = True)
    """
    if gis is None:
        gis = arcgis.env.active_gis
    if isinstance(length_unit, LengthUnits):
        length_unit = length_unit.value
    if isinstance(area_unit, AreaUnits):
        area_unit = area_unit.value
    return gis._tools.geometry.areas_and_lengths(
        polygons,
        length_unit,
        area_unit,
        calculation_type,
        spatial_ref,
        future=future,
    )


# -------------------------------------------------------------------------
def auto_complete(
    polygons: Optional[list[Polygon]] = None,
    polylines: Optional[list[Polyline]] = None,
    spatial_ref: Optional[SpatialReference] = None,
    gis: Optional[GIS] = None,
    future: bool = False,
):
    """
    The ``auto_complete`` function simplifies the process of
    constructing new :class:`~arcgis.geometry.Polygon` objects that are adjacent to other polygons.
    It constructs polygons that fill in the gaps between existing
    polygons and a set of :class:`~arcgis.geometry.Polyline` objects.

    ================  ===============================================================================
    **Keys**          **Description**
    ----------------  -------------------------------------------------------------------------------
    polygons          A List of :class:`~arcgis.geometry.Polygon` objects
    ----------------  -------------------------------------------------------------------------------
    polylines         A List of :class:`~arcgis.geometry.Polyline` objects
    ----------------  -------------------------------------------------------------------------------
    spatial_ref       A :class:`~arcgis.geometry.SpatialReference` of the input geometries WKID
    ----------------  -------------------------------------------------------------------------------
    future            Optional boolean. If True, a future object will be returned and the process
                      will not wait for the task to complete. The default is False, which means wait for results.
    ================  ===============================================================================

    :returns:
        A :class:`~arcgis.geometry.Polygon` object, or a `GeometryJob` object. If ``future = True``,
        then the result is a :class:`~concurrent.futures.Future` object. Call ``result()`` to get the response.
    """
    if gis is None:
        gis = arcgis.env.active_gis
    return gis._tools.geometry.auto_complete(
        polygons, polylines, spatial_ref, future=future
    )


def buffer(
    geometries: list,
    in_sr: Union[int, dict[str, Any]],
    distances: float | list[float],
    unit: str | LengthUnits,
    out_sr: Optional[Union[int, dict[str, Any]]] = None,
    buffer_sr: Optional[float] = None,
    union_results: Optional[bool] = None,
    geodesic: Optional[bool] = None,
    gis: Optional[GIS] = None,
    future: bool = False,
):
    """
    The ``buffer`` function is performed on a geometry service resource
    The result of this function is a buffered :class:`~arcgis.geometry.Polygon` at the
    specified distances for the input :class:`~arcgis.geometry.Geometry` array.

    .. note::
        The options are available to union buffers and to use geodesic distance.

    ================  ===============================================================================
    **Keys**          **Description**
    ----------------  -------------------------------------------------------------------------------
    geometries        The array of geometries to be buffered
    ----------------  -------------------------------------------------------------------------------
    in_sr             The well-known ID of the :class:`~arcgis.geometry.SpatialReference` or a spatial
                      reference JSON object for the input geometries.
    ----------------  -------------------------------------------------------------------------------
    distances         The distances that each of the input geometries is
                      buffered.
    ----------------  -------------------------------------------------------------------------------
    unit              The units for calculating each buffer distance. If unit
                      is not specified, the units are derived from ``bufferSR``. If
                      ``bufferSR`` is not specified, the units are derived from ``in_sr``.
    ----------------  -------------------------------------------------------------------------------
    out_sr            The well-known ID of the :class:`~arcgis.geometry.SpatialReference` or a
                      spatial reference JSON object for the output geometries.
    ----------------  -------------------------------------------------------------------------------
    buffer_sr         The well-known ID of the :class:`~arcgis.geometry.SpatialReference` or a
                      spatial reference JSON object for the buffer geometries.
    ----------------  -------------------------------------------------------------------------------
    union_results     A boolean. If True, all geometries buffered at a given
                      distance are unioned into a single (gis,possibly multipart)
                      :class:`~arcgis.geometry.Polygon`, and the unioned geometry is placed in the output
                      array. The default is False.
    ----------------  -------------------------------------------------------------------------------
    geodesic          Set geodesic to true to buffer the input geometries
                      using geodesic distance. Geodesic distance is the shortest
                      path between two points along the ellipsoid of the earth. If
                      geodesic is set to False, the 2D Euclidean distance is used
                      to buffer the input geometries.

                      .. note::
                        The default value depends on the `geometry type`, `unit` and `bufferSR`.
    ----------------  -------------------------------------------------------------------------------
    future            Optional boolean. If True, a future object will be returned and the process
                      will not wait for the task to complete. The default is False, which means wait for results.
                      If setting future to True there is a limitation of 6500 geometries that can be processed in one call.
    ================  ===============================================================================

    :returns:
        A list of :class:`~arcgis.geometry.Polygon` object, or a `GeometryJob` object. If ``future = True``,
        then the result is a :class:`~concurrent.futures.Future` object. Call ``result()`` to get the response.

    .. code-block:: python

            >>> buffer(geometries =[geom1, geom2,...],
                       in_sr = "wkid_in",
                       unit = LengthUnits.METER,
                       out_sr = "wkid_out",
                       buffer_sr = "wkid_buffer",
                       union_results =True,
                       geodesic = True,
                       future = True)

    """
    if gis is None:
        gis = arcgis.env.active_gis
    if isinstance(unit, LengthUnits):
        unit = unit.value
    if isinstance(distances, list):
        distances = ",".join([str(d) for d in distances])
    return gis._tools.geometry.buffer(
        geometries,
        in_sr,
        distances,
        unit,
        out_sr,
        buffer_sr,
        union_results,
        geodesic,
        future=future,
    )


def convex_hull(
    geometries: Union[list[Polygon], list[Polyline], list[MultiPoint], list[Point]],
    spatial_ref: Optional[Union[int, dict[str, Any]]] = None,
    gis: Optional[GIS] = None,
    future: bool = False,
):
    """
    The `convex_hull` function is performed on a `Geometry Service
    resource <https://developers.arcgis.com/rest/services-reference/enterprise/geometry-service.htm>`_.
    It returns the minimum bounding shape that contains the input geometry. The
    input geometry can be a :class:`~arcgis.geometry.Point`, :class:`~arcgis.geometry.MultiPoint`,
    :class:`~arcgis.geometry.Polyline` , or :class:`~arcgis.geometry.Polygon` object.

    .. note::
        The convex hull is typically a polygon but can also be a polyline
        or point in degenerate cases.

    ================  ===============================================================================
    **Keys**          **Description**
    ----------------  -------------------------------------------------------------------------------
    geometries        A list of :class:`~arcgis.geometry.Point`, :class:`~arcgis.geometry.MultiPoint`,
                      :class:`~arcgis.geometry.Polyline`, or :class:`~arcgis.geometry.Polygon` objects.
                      The structure of each geometry in the array is defined the same as the
                      `JSON geometry objects <https://developers.arcgis.com/documentation/common-data-types/geometry-objects.htm>`_
                      returned by the ArcGIS REST API.

                      .. note::
                          :class:`~arcgis.geometry.Geometry` objects can be obtained by querying a
                          :class:`~arcgis.features.FeatureLayer`, returning it as a Pandas
                          data frame, and then assigning variables to a geometry based on the row index.

                          .. code-block:: python

                              >>> flyr_item = gis.content.search("*", "Feature Layer")[0]

                              >>> flyr_df = flyr_item.query(where="1=1", as_df=True)
                              >>> geom0 = flyr_df.loc[0].SHAPE

    ----------------  -------------------------------------------------------------------------------
    spatial_ref       An integer value, or a :class:`~arcgis.geometry.SpatialReference` object
                      defined using the the Well-Known ID (`wkid`) of the Spatial Reference.

                      .. note:: See `Spatial Reference <https://developers.arcgis.com/documentation/common-data-types/geometry-objects.htm#GUID-DFF0E738-5A42-40BC-A811-ACCB5814BABC>`_
                          in the `Geometry objects` help, or `Using Spatial References <https://developers.arcgis.com/rest/services-reference/enterprise/using-spatial-references.htm>`_
                          for details on concepts and resources for finding specific `wkid` values.

                      .. code-block:: python

                          >>> geom_result = convex_hull(geometries=[geometry_object]
                                                        spatial_ref=<wkid>)

                      or

                      .. code-block:: python

                          >>> geom_result = convex_hull(geometries=[geometry_object],
                                                        spatial_ref={"wkid": <wkid>})

                      or

                      .. code-block:: python

                          >>> from arcgis.geometry import SpatialReference
                          >>> sr_obj_wkid = SpatialReference(<wkid>)

                          >>> geom_result = convex_hull(geometries=[geometry_object],
                                                        spatial_ref=sr_obj_wkid)

    ----------------  -------------------------------------------------------------------------------
    future            Optional boolean. If True, a future object will be returned and the process
                      will not wait for the task to complete. The default is False, which means wait for results.
                      If setting future to True there is a limitation of 6500 geometries that can be processed in one call.
    ================  ===============================================================================

    :returns:
        A list containing the :class:`~arcgis.geometry.Geometry` object of the result, or  if ``future=True``,
        a :class:`~concurrent.futures.Future` object. Call ``result()`` on the `future` to get
        the response details.


    .. code-block:: python

        # Usage Example:

        >>> from arcgis.gis import GIS
        >>> from arcgis.geometry import convex_hull

        >>> gis = GIS(profile="your_organization_profile")

        >>> flyr_item = gis.content.get("<item_id for feature layer>")
        >>> flyr = flyr_item.layers[0]

        >>> df = flyr.query(where="OBJECTID=1", as_df=True)

        >>> geom1 = df.loc[0].SHAPE
        >>> hull_geom1 = convex_hull(geometries=[geom1],
                                     spatial_ref={"wkid": 2056})

        >>> hull_geom1[0]

        {'rings': [[[2664507.7925999984, 1212609.7138999999],
        .,
        .,
        [2664678.264199998, 1212618.6860999987],
        [2664507.7925999984, 1212609.7138999999]]],
        'spatialReference': {'wkid': {'wkid': 2056}}}
    """
    if gis is None:
        gis = arcgis.env.active_gis
    return gis._tools.geometry.convex_hull(geometries, spatial_ref, future=future)


def cut(
    cutter: Polyline,
    target: Union[list[Polyline], list[Polygon]],
    spatial_ref: Optional[Union[int, dict[str, Any]]] = None,
    gis: Optional[GIS] = None,
    future: bool = False,
):
    """
    The `cut` function is performed on a :class:`~arcgis.geometry.Geometry` service resource. This
    function splits the target :class:`~arcgis.geometry.Polyline` or :class:`~arcgis.geometry.Polygon` where it is
    crossed by the cutter polyline.

    .. note::
        At 10.1 and later, this function calls simplify on the input
        cutter and target geometries.

    ================  ===============================================================================
    **Keys**          **Description**
    ----------------  -------------------------------------------------------------------------------
    cutter            The :class:`~arcgis.geometry.Polyline` that will be used to divide the target
                      into pieces where it crosses the target.The spatial reference
                      of the polylines is specified by ``spatial_ref``.

                      .. note::
                        The structure of the
                        polyline is the same as the structure of the JSON polyline
                        objects returned by the ArcGIS REST API.
    ----------------  -------------------------------------------------------------------------------
    target            The array of :class:`~arcgis.geometry.Polyline` or :class:`~arcgis.geometry.Polygon` to be cut.
                      The structure of the geometry is the same as the structure of the
                      JSON geometry objects returned by the ArcGIS REST API. The
                      spatial reference of the target geometry array is specified by
                      spatial_ref.
    ----------------  -------------------------------------------------------------------------------
    spatial_ref       A :class:`~arcgis.geometry.SpatialReference` of the input geometries Well-Known ID or a JSON
                      object for the output geometry
    ----------------  -------------------------------------------------------------------------------
    future            Optional boolean. If True, a future object will be returned and the process
                      will not wait for the task to complete. The default is False, which means wait for results.
                      If setting future to True there is a limitation of 6500 geometries that can be processed in one call.
    ================  ===============================================================================

    :returns:
        A List of :class:`~arcgis.geometry.Geometry` objects, or a `GeometryJob` object. If ``future = True``,
        then the result is a :class:`~concurrent.futures.Future` object. Call ``result()`` to get the response.
    """
    if gis is None:
        gis = arcgis.env.active_gis
    return gis._tools.geometry.cut(cutter, target, spatial_ref, future=future)


def densify(
    geometries: Union[list[Polygon], list[Polyline], list[MultiPoint], list[Point]],
    spatial_ref: Optional[Union[int, dict[str, Any]]],
    max_segment_length: Optional[float],
    length_unit: Optional[str] | Optional[LengthUnits],
    geodesic: bool = False,
    gis: Optional[GIS] = None,
    future: bool = False,
):
    """
    The ``densify`` function is performed using the :class:`~arcgis.gis.GIS` geometry engine.
    This function densifies :class:`~arcgis.geometry.Geometry` objects by plotting :class:`~arcgis.geometry.Point`
    objects between existing vertices.


    ================  ===============================================================================
    **Keys**          **Description**
    ----------------  -------------------------------------------------------------------------------
    geometries        An array of :class:`~arcgis.geometry.Point`, :class:`~arcgis.geometry.MultiPoint`,
                      :class:`~arcgis.geometry.Polyline`, or :class:`~arcgis.geometry.Polygon` objects.
                      The structure of each geometry in the array is the
                      same as the structure of the JSON geometry objects returned by
                      the ArcGIS REST API.
    ----------------  -------------------------------------------------------------------------------
    spatial_ref       The ``well-known ID`` or a spatial reference JSON object for
                      the input :class:`~arcgis.geometry.Polyline` object.

                      .. note::
                        For a list of valid WKID values, see
                        Projected coordinate systems and Geographic coordinate systems.
    ----------------  -------------------------------------------------------------------------------
    max_segment_len   All segments longer than ``maxSegmentLength`` are
                      replaced with sequences of lines no longer than ``max_segment_length``.
    ----------------  -------------------------------------------------------------------------------
    length_unit       The length unit of ``max_segment_length``. If ``geodesic`` is
                      set to `false`, then the units are derived from ``spatial_ref``, and
                      ``length_unit`` is ignored. If ``geodesic`` is set to `true`, then
                      ``length_unit`` must be a linear unit. In a case where ``length_unit`` is
                      not specified and ``spatial_ref`` is a PCS, the units are derived from ``spatial_ref``.
                      In a case where ``length_unit`` is not specified and ``spatial_ref`` is a GCS,
                      then the units are meters.
    ----------------  -------------------------------------------------------------------------------
    geodesic          If geodesic is set to true, then geodesic distance is
                      used to calculate max_segment_length. Geodesic distance is the
                      shortest path between two points along the ellipsoid of the
                      earth. If geodesic is set to false, then 2D Euclidean distance
                      is used to calculate max_segment_length. The default is false.
    ----------------  -------------------------------------------------------------------------------
    future            Optional boolean. If True, a future object will be returned and the process
                      will not wait for the task to complete. The default is False, which means wait for results.
                      If setting future to True there is a limitation of 6500 geometries that can be processed in one call.
    ================  ===============================================================================

    :returns:
        A list of :class:`~arcgis.geometry.Geometry` object, or a `GeometryJob` object. If ``future = True``,
        then the result is a :class:`~concurrent.futures.Future` object. Call ``result()`` to get the response.

    .. code-block:: python

            >>> densify(geometries =[geom1, geom2,...],
                        spatial_ref = "wkid",
                        max_segment_length = 100.0,
                        length_unit = LengthUnits.METER,
                        geodesic = True,
                        future = False)

    """
    if gis is None:
        gis = arcgis.env.active_gis
    if isinstance(length_unit, LengthUnits):
        length_unit = length_unit.value

    return gis._tools.geometry.densify(
        geometries,
        spatial_ref,
        max_segment_length,
        length_unit,
        geodesic,
        future=future,
    )


def difference(
    geometries: Union[list[Polygon], list[Polyline], list[MultiPoint], list[Point]],
    spatial_ref: Optional[Union[int, dict[str, Any]]],
    geometry: Geometry,
    gis: Optional[GIS] = None,
    future: bool = False,
):
    """
    The ``difference`` function is performed on a geometry service
    resource. This function constructs the set-theoretic difference
    between each element of an array of geometries and another geometry
    the so-called difference geometry. In other words, let B be the
    difference geometry. For each geometry, A, in the input geometry
    array, it constructs A-B.

    ================  ===============================================================================
    **Keys**          **Description**
    ----------------  -------------------------------------------------------------------------------
    geometries        An array of :class:`~arcgis.geometry.Point`, :class:`~arcgis.geometry.MultiPoint`,
                      :class:`~arcgis.geometry.Polyline`, or :class:`~arcgis.geometry.Polygon` objects.
                      The structure of each geometry in the array is the
                      same as the structure of the JSON geometry objects returned by
                      the ArcGIS REST API.
    ----------------  -------------------------------------------------------------------------------
    geometry          A single geometry of any type and of a dimension equal
                      to or greater than the elements of geometries. The structure of
                      geometry is the same as the structure of the JSON geometry
                      objects returned by the ArcGIS REST API. The use of simple
                      syntax is not supported.
    ----------------  -------------------------------------------------------------------------------
    spatial_ref       A :class:`~arcgis.geometry.SpatialReference` of the input geometries Well-Known ID or JSON object
    ----------------  -------------------------------------------------------------------------------
    future            Optional boolean. If True, a future object will be returned and the process
                      will not wait for the task to complete. The default is False, which means wait for results.
                      If setting future to True there is a limitation of 6500 geometries that can be processed in one call.
    ================  ===============================================================================

    :returns:
        A list of :class:`~arcgis.geometry.Geometry` objects, or a `GeometryJob` object. If ``future = True``,
        then the result is a :class:`~concurrent.futures.Future` object. Call ``result()`` to get the response.
    """
    if gis is None:
        gis = arcgis.env.active_gis

    return gis._tools.geometry.difference(
        geometries, spatial_ref, geometry, future=future
    )


def distance(
    spatial_ref: Optional[Union[int, dict[str, Any]]],
    geometry1: Geometry,
    geometry2: Geometry,
    distance_unit: str | LengthUnits | None = "",
    geodesic: bool = False,
    gis: Optional[GIS] = None,
    future: bool = False,
):
    """
    The ``distance`` function is performed on a geometry service resource.
    It reports the `2D Euclidean` or `geodesic` distance between the two
    :class:`~arcgis.geometry.Geometry` objects.

    ================  ===============================================================================
    **Keys**          **Description**
    ----------------  -------------------------------------------------------------------------------
    geometry1         The :class:`~arcgis.geometry.Geometry` object from which the distance is
                      measured. The structure of each geometry in the array is the
                      same as the structure of the JSON geometry objects returned by
                      the ArcGIS REST API.
    ----------------  -------------------------------------------------------------------------------
    geometry2         The :class:`~arcgis.geometry.Geometry` object to which the distance is
                      measured. The structure of each geometry in the array is the
                      same as the structure of the JSON geometry objects returned by
                      the ArcGIS REST API.
    ----------------  -------------------------------------------------------------------------------
    distance_unit     Optional. One of :class:`~arcgis.geometry.functions.LengthUnits` enumeration
                      members. See Geometry Service
                      `distance <https://developers.arcgis.com/rest/services-reference/enterprise/distance.htm>`_
                      for full details.
    ----------------  -------------------------------------------------------------------------------
    geodesic          If ``geodesic`` is set to true, then the geodesic distance
                      between the ``geometry1`` and ``geometry2`` geometries is returned.
                      Geodesic distance is the shortest path between two points along
                      the ellipsoid of the earth. If ``geodesic`` is set to false or not
                      specified, the planar distance is returned. The default value is false.
    ----------------  -------------------------------------------------------------------------------
    spatial_ref       A :class:`~arcgis.geometry.SpatialReference` of the input geometries Well-Known
                      ID or JSON object
    ----------------  -------------------------------------------------------------------------------
    future            Optional boolean. If True, a future object will be returned and the process
                      will not wait for the task to complete. The default is False, which means wait for results.
    ================  ===============================================================================

    :returns:
        The 2D or geodesic distance between the two :class:`~arcgis.geometry.Geometry` objects, or
        a `GeometryJob` object. If ``future = True``,
        then the result is a :class:`~concurrent.futures.Future` object. Call ``result()`` to get the response.
    """
    if gis is None:
        gis = arcgis.env.active_gis
    if isinstance(distance_unit, LengthUnits):
        distance_unit = distance_unit.value
    elif distance_unit in [None, ""]:
        distance_unit = ""
    return gis._tools.geometry.distance(
        spatial_ref,
        geometry1,
        geometry2,
        distance_unit,
        geodesic,
        future=future,
    )


def find_transformation(
    in_sr: Optional[Union[int, dict[str, Any]]],
    out_sr: Optional[Union[int, dict[str, Any]]],
    extent_of_interest: Optional[dict[str, Any]] = None,
    num_of_results: int = 1,
    gis: Optional[GIS] = None,
    future: bool = False,
):
    """
    The ``find_transformations`` function is performed on a :class:`~arcgis.geometry.Geometry`
    service resource. This function returns a list of applicable
    geographic transformations you should use when projecting
    geometries from the input :class:`~arcgis.geometry.SpatialReference` to the output
    :class:`~arcgis.geometry.SpatialReference`. The transformations are in JSON format and are returned
    in order of most applicable to least applicable. Recall that a
    geographic transformation is not needed when the input and output
    spatial references have the same underlying geographic coordinate
    systems. In this case, findTransformations returns an empty list.

    .. note::
        Every returned geographic transformation is a forward
        transformation meaning that it can be used as-is to project from
        the input spatial reference to the output spatial reference. In the
        case where a predefined transformation needs to be applied in the
        reverse direction, it is returned as a forward composite
        transformation containing one transformation and a transformForward
        element with a value of false.

    ================  ===============================================================================
    **Keys**          **Description**
    ----------------  -------------------------------------------------------------------------------
    in_sr             The well-known ID of the :class:`~arcgis.geometry.SpatialReference` or a spatial
                      reference JSON object for the input geometries.
    ----------------  -------------------------------------------------------------------------------
    out_sr            The well-known ID of the :class:`~arcgis.geometry.SpatialReference` or a
                      spatial reference JSON object for the output geometries.
    ----------------  -------------------------------------------------------------------------------
    ext_of_interest   The bounding box of the area of interest specified as a JSON envelope.If provided, the extent of
                      interest is used to return the most applicable geographic
                      transformations for the area.

                      .. note::
                        If a :class:`~arcgis.geometry.SpatialReference` is not
                        included in the JSON envelope, the ``in_sr`` is used for the
                        envelope.

    ----------------  -------------------------------------------------------------------------------
    num_of_results    The number of geographic transformations to
                      return. The default value is 1.

                      .. note::
                        If ``num_of_results`` has a value of -1, all applicable transformations are returned.
    ----------------  -------------------------------------------------------------------------------
    future            Optional boolean. If True, a future object will be returned and the process
                      will not wait for the task to complete. The default is False, which means wait for results.
    ================  ===============================================================================

    :returns:
        A List of geographic transformations, or a `GeometryJob` object. If ``future = True``,
        then the result is a :class:`~concurrent.futures.Future` object. Call ``result()`` to get the response.
    """
    if gis is None:
        gis = arcgis.env.active_gis
    return gis._tools.geometry.find_transformation(
        in_sr, out_sr, extent_of_interest, num_of_results, future=future
    )


def from_geo_coordinate_string(
    spatial_ref: Optional[Union[int, dict[str, Any]]],
    strings: list[str],
    conversion_type: Optional[str],
    conversion_mode: Optional[str] = None,
    gis: Optional[GIS] = None,
    future: bool = False,
):
    """
    The ``from_geo_coordinate_string`` function is performed on a :class:`~arcgis.geometry.Geometry`
    service resource. The function converts an array of well-known
    strings into xy-coordinates based on the conversion type and
    :class:`~arcgis.geometry.SpatialReference` supplied by the user. An optional conversion mode
    parameter is available for some conversion types. See :attr:`~arcgis.geometry.functions.to_geo_coordinate_strings`
    for more information on the opposite conversion.


    ================  ===============================================================================
    **Keys**          **Description**
    ----------------  -------------------------------------------------------------------------------
    spatial_ref       A :class:`~arcgis.geometry.SpatialReference` of the input geometries Well-Known ID or JSON object
    ----------------  -------------------------------------------------------------------------------
    strings           An array of strings formatted as specified by conversion_type.
                      Syntax: [<string1>,...,<stringN>]
    ----------------  -------------------------------------------------------------------------------
    conversion-type   The conversion type of the input strings.

                      .. note::
                        Valid conversion types are:
                        `MGRS` - Military Grid Reference System
                        `USNG` - United States National Grid
                        `UTM` - Universal Transverse Mercator
                        `GeoRef` - World Geographic Reference System
                        `GARS` - Global Area Reference System
                        `DMS` - Degree Minute Second
                        `DDM` - Degree Decimal Minute
                        `DD` - Decimal Degree
    ----------------  -------------------------------------------------------------------------------
    conversion_mode   Conversion options for MGRS, UTM and GARS conversion types.

                      .. note::
                        Valid conversion modes for MGRS are:
                        `mgrsDefault` - Default. Uses the spheroid from the given spatial reference.

                        `mgrsNewStyle` - Treats all spheroids as new, like WGS 1984. The 80 degree longitude falls into Zone 60.

                        `mgrsOldStyle` - Treats all spheroids as old, like Bessel 1841. The 180 degree longitude falls into Zone 60.

                        `mgrsNewWith180InZone01` - Same as mgrsNewStyle except the 180 degree longitude falls into Zone 01

                        `mgrsOldWith180InZone01` - Same as mgrsOldStyle except the 180 degree longitude falls into Zone 01

                      .. note::
                        Valid conversion modes for UTM are:
                        `utmDefault` - Default. No options.
                        `utmNorthSouth` - Uses north/south latitude indicators instead of
                        `zone numbers` - Non-standard. Default is recommended

    ----------------  -------------------------------------------------------------------------------
    future            Optional boolean. If True, a future object will be returned and the process
                      will not wait for the task to complete. The default is False, which means wait for results.
    ================  ===============================================================================

    :returns:
        An array of (x,y) coordinates, or a `GeometryJob` object. If ``future = True``,
        then the result is a :class:`~concurrent.futures.Future` object. Call ``result()`` to get the response.

    .. code-block:: python

            >>> coords = from_geo_coordinate_string(spatial_ref = "wkid",
                                            strings = ["01N AA 66021 00000","11S NT 00000 62155", "31U BT 94071 65288"]
                                            conversion_type = "MGRS",
                                            conversion_mode = "mgrs_default",
                                            future = False)
            >>> coords
                [[x1,y1], [x2,y2], [x3,y3]]
    """
    if gis is None:
        gis = arcgis.env.active_gis
    return gis._tools.geometry.from_geo_coordinate_string(
        spatial_ref, strings, conversion_type, conversion_mode, future=future
    )


def generalize(
    spatial_ref: Optional[Union[int, dict[str, Any]]],
    geometries: list[Geometry],
    max_deviation: int,
    deviation_unit: str | LengthUnits | None = None,
    gis: Optional[GIS] = None,
    future: bool = False,
):
    """
    The ``generalize`` function is performed on a :class:`~arcgis.geometry.Geometry` service
    resource. The `generalize` function simplifies the input geometries
    using the `Douglas-Peucker` algorithm with a specified maximum
    deviation distance.

    .. note::
        The output geometries will contain a subset of
        the original input vertices.


    ================  ===============================================================================
    **Keys**          **Description**
    ----------------  -------------------------------------------------------------------------------
    geometries        The array :class:`~arcgis.geometry.Geometry` objects to be generalized.
    ----------------  -------------------------------------------------------------------------------
    max_deviation     ``max_deviation`` sets the maximum allowable offset,
                      which will determine the degree of simplification. This value
                      limits the distance the output geometry can differ from the input
                      geometry.
    ----------------  -------------------------------------------------------------------------------
    deviation_unit    If ``geodesic`` is set to true, then the geodesic distance
                      between the ``geometry1`` and ``geometry2`` geometries is returned.
                      Geodesic distance is the shortest path between two points along
                      the ellipsoid of the earth. If ``geodesic`` is set to false or not
                      specified, the planar distance is returned. The default value is false.
    ----------------  -------------------------------------------------------------------------------
    spatial_ref       A :class:`~arcgis.geometry.SpatialReference` of the input geometries Well-Known ID or JSON object
    ----------------  -------------------------------------------------------------------------------
    future            Optional boolean. If True, a future object will be returned and the process
                      will not wait for the task to complete. The default is False, which means wait for results.
                      If setting future to True there is a limitation of 6500 geometries that can be processed in one call.
    ================  ===============================================================================

    :returns:
        An array of the simplified :class:`~arcgis.geometry.Geometry` objects, or a `GeometryJob` object. If ``future = True``,
        then the result is a :class:`~concurrent.futures.Future` object. Call ``result()`` to get the response.
    """
    if gis is None:
        gis = arcgis.env.active_gis
    if isinstance(deviation_unit, LengthUnits):
        deviation_unit = deviation_unit.value
    elif deviation_unit is None:
        deviation_unit = ""
    return gis._tools.geometry.generalize(
        spatial_ref, geometries, max_deviation, deviation_unit, future=future
    )


def intersect(
    spatial_ref: Optional[Union[int, dict[str, Any]]],
    geometries: list[Geometry],
    geometry: Geometry,
    gis: Optional[GIS] = None,
    future: bool = False,
):
    """
    The ``intersect`` function is performed on a :class:`~arcgis.geometry.Geometry` service
    resource. This function constructs the set-theoretic intersection
    between an array of geometries and another geometry.

    .. note::
        The dimension of each resultant geometry is the minimum dimension of the input
        geometry in the geometries array and the other geometry specified by the geometry parameter.


    ================  ===============================================================================
    **Keys**          **Description**
    ----------------  -------------------------------------------------------------------------------
    geometries        An array of :class:`~arcgis.geometry.Point`, :class:`~arcgis.geometry.MultiPoint`,
                      :class:`~arcgis.geometry.Polyline`, or :class:`~arcgis.geometry.Polygon` objects.
                      The structure of each geometry in the array is the
                      same as the structure of the JSON geometry objects returned by
                      the ArcGIS REST API.
    ----------------  -------------------------------------------------------------------------------
    geometry          A single :class:`~arcgis.geometry.Geometry` of any type and of a dimension equal
                      to or greater than the elements of geometries. The structure of
                      geometry is the same as the structure of the JSON geometry
                      objects returned by the ArcGIS REST API. The use of simple
                      syntax is not supported.
    ----------------  -------------------------------------------------------------------------------
    spatial_ref       A :class:`~arcgis.geometry.SpatialReference` of the input geometries Well-Known ID or JSON object
    ----------------  -------------------------------------------------------------------------------
    future            Optional boolean. If True, a future object will be returned and the process
                      will not wait for the task to complete. The default is False, which means wait for results.
                      If setting future to True there is a limitation of 6500 geometries that can be processed in one call.
    ================  ===============================================================================

    :returns:
        The set-theoretic dimension between :class:`~arcgis.geometry.Geometry` objects, or a `GeometryJob` object. If ``future = True``,
        then the result is a :class:`~concurrent.futures.Future` object. Call ``result()`` to get the response.
    """
    if gis is None:
        gis = arcgis.env.active_gis
    return gis._tools.geometry.intersect(
        spatial_ref, geometries, geometry, future=future
    )


def label_points(
    spatial_ref: Optional[Union[int, dict[str, Any]]],
    polygons: list[Polygon],
    gis: Optional[GIS] = None,
    future: bool = False,
):
    """
    The ``label_points`` function is performed on a :class:`~arcgis.geometry.Geometry` service
    resource. The ``labelPoints`` function calculates an interior :class:`~arcgis.geometry.Point`
    for each :class:`~arcgis.geometry.Polygon` specified in the input array. These interior
    points can be used by clients for labeling the polygons.


    ================  ===============================================================================
    **Keys**          **Description**
    ----------------  -------------------------------------------------------------------------------
    polygons          An array of :class:`~arcgis.geometry.Polygon` objects whose label :class:`~arcgis.geometry.Point`
                      objects are to be computed. The spatial reference of the polygons is specified by
                      ``spatial_ref``.
    ----------------  -------------------------------------------------------------------------------
    spatial_ref       A :class:`~arcgis.geometry.SpatialReference` of the input geometries Well-Known ID or JSON object
    ----------------  -------------------------------------------------------------------------------
    future            Optional boolean. If True, a future object will be returned and the process
                      will not wait for the task to complete. The default is False, which means wait for results.
    ================  ===============================================================================

    :returns:
        An array of :class:`~arcgis.geometry.Point` objects, or a `GeometryJob` object. If ``future = True``,
        then the result is a :class:`~concurrent.futures.Future` object. Call ``result()`` to get the response.
    """
    if gis is None:
        gis = arcgis.env.active_gis
    return gis._tools.geometry.label_points(spatial_ref, polygons, future=future)


def lengths(
    spatial_ref: Optional[Union[int, dict[str, Any]]],
    polylines: Polyline,
    length_unit: str | LengthUnits,
    calculation_type: str,
    gis: Optional[GIS] = None,
    future: bool = False,
):
    """
    The ``lengths`` function is performed on a :class:`~arcgis.geometry.Geometry` service resource.
    This function calculates the` 2D Euclidean` or `geodesic` lengths of
    each :class:`~arcgis.geometry.Polyline` specified in the input array.

    ================  ===============================================================================
    **Keys**          **Description**
    ----------------  -------------------------------------------------------------------------------
    polylines         The array of :class:`~arcgis.geometry.Polyline` whose lengths are to be computed.
    ----------------  -------------------------------------------------------------------------------
    length_unit       The length unit in which the length of
                      :class:`~arcgis.geometry.Polyline` will be calculated. If ``calculation_type``
                      is planar, then ``length_unit`` can be any `esriUnits`
                      constant. If ``lengthUnit`` is not specified, the
                      units are derived from ``spatial_ref``. If ``calculationType`` is
                      not planar, then `lengthUnit` must be a linear
                      esriUnits constant, such as `esriSRUnit_Meter` or
                      `esriSRUnit_SurveyMile`. If ``length_unit`` is not
                      specified, the units are meters. For a list of
                      valid units, see `esriSRUnitType Constants` and
                      `esriSRUnit2Type Constant`.
    ----------------  -------------------------------------------------------------------------------
    calculation_type  The type defined for the length calculation of the input geometries. The type can be one
                      of the following values:

                          1. planar - Planar measurements use 2D Euclidean distance to calculate area and length. This
                          should only be used if the area or length needs to be calculated in the given
                          :class:`~arcgis.geometry.SpatialReference`. Otherwise, use ``preserveShape``.

                          2. geodesic - Use this type if you want to calculate an area or length using only the vertices
                          of the :class:`~arcgis.geometry.Polygon` and define the lines between the points as geodesic
                          segments independent of the actual shape of the :class:`~arcgis.geometry.Polygon`. A geodesic
                          segment is the shortest path between two points on an ellipsoid.

                          3. preserveShape - This type calculates the area or length of the geometry on the surface of
                          the Earth ellipsoid. The shape of the geometry in its coordinate system is preserved.
    ----------------  -------------------------------------------------------------------------------
     future           Optional boolean. If True, a future object will be returned and the process
                      will not wait for the task to complete. The default is False, which means wait for results.
    ================  ===============================================================================

    :returns:
        A list of floats of 2D-Euclidean or Geodesic lengths, or a `GeometryJob` object. If ``future = True``,
        then the result is a :class:`~concurrent.futures.Future` object. Call ``result()`` to get the response.
    """
    if gis is None:
        gis = arcgis.env.active_gis
    if isinstance(length_unit, LengthUnits):
        length_unit = length_unit.value
    service = gis._tools.geometry
    return service.lengths(
        spatial_ref, polylines, length_unit, calculation_type, future=future
    )


def offset(
    geometries: Union[list[Polygon], list[Polyline], list[MultiPoint], list[Point]],
    offset_distance: float,
    offset_unit: str | LengthUnits,
    offset_how: str = "esriGeometryOffsetRounded",
    bevel_ratio: int = 10,
    simplify_result: bool = False,
    spatial_ref: Optional[Union[int, dict[str, Any]]] = None,
    gis: Optional[GIS] = None,
    future: bool = False,
):
    """
    The ``offset`` function is performed on a :class:`~arcgis.geometry.Geometry` service resource.
    This function constructs geometries that are offset from the
    given input geometries. If the offset parameter is positive, the
    constructed offset will be on the right side of the geometry. Left
    side offsets are constructed with negative parameters.

    .. note::
        Tracing the geometry from its first vertex to the last will give you a
        direction along the geometry. It is to the right and left
        perspective of this direction that the positive and negative
        parameters will dictate where the offset is constructed. In these
        terms, it is simple to infer where the offset of even horizontal geometries will be constructed.

    ================  ===============================================================================
    **Keys**          **Description**
    ----------------  -------------------------------------------------------------------------------
    geometries        An array of :class:`~arcgis.geometry.Point`, :class:`~arcgis.geometry.MultiPoint`,
                      :class:`~arcgis.geometry.Polyline`, or :class:`~arcgis.geometry.Polygon` objects.
                      The structure of each geometry in the array is the
                      same as the structure of the JSON geometry objects returned by
                      the ArcGIS REST API.
    ----------------  -------------------------------------------------------------------------------
    offset_distance   Specifies the distance for constructing an offset
                      based on the input geometries.

                      .. note::
                        If the ``offset_distance`` parameter is
                        positive, the constructed offset will be on the right side of the
                        curve. Left-side offsets are constructed with negative values.
    ----------------  -------------------------------------------------------------------------------
    offset_unit       A unit for offset distance. If a unit is not specified, the units are derived from
                      ``spatial_ref``.
    ----------------  -------------------------------------------------------------------------------
    offset_how        The ``offset_how`` parameter determines how outer corners between segments are handled.
                      The three options are as follows:
                        1. ``esriGeometryOffsetRounded`` - Rounds the corner between extended offsets.
                        2. ``esriGeometryOffsetBevelled`` - Squares off the corner after a given ratio distance.
                        3. ``esriGeometryOffsetMitered`` - Attempts to allow extended offsets to naturally intersect,
                        but if that intersection occurs too far from the corner, the corner is eventually bevelled off
                        at a fixed distance.
    ----------------  -------------------------------------------------------------------------------
    bevel_ratio       ``bevel_ratio`` is multiplied by the ``offset_distance``, and
                      the result determines how far a mitered offset intersection can
                      be located before it is bevelled. When mitered is specified,
                      bevel_ratio is ignored and 10 is used internally. When bevelled is
                      specified, 1.1 will be used if bevel_ratio is not specified.
                      ``bevel_ratio`` is ignored for rounded offset.
    ----------------  -------------------------------------------------------------------------------
    simplify_result   if ``simplify_result`` is set to true, then self
                      intersecting loops will be removed from the result offset geometries. The default is false.
    ----------------  -------------------------------------------------------------------------------
    spatial_ref       A :class:`~arcgis.geometry.SpatialReference` of the input geometries Well-Known ID or JSON object
    ----------------  -------------------------------------------------------------------------------
    future            Optional boolean. If True, a future object will be returned and the process
                      will not wait for the task to complete. The default is False, which means wait for results.
                      If setting future to True there is a limitation of 6500 geometries that can be processed in one call.
    ================  ===============================================================================

    :returns:
        A list of :class:`~arcgis.geometry.Geometry` objects, or a `GeometryJob` object. If ``future = True``,
        then the result is a :class:`~concurrent.futures.Future` object. Call ``result()`` to get the response.

    .. code-block:: python

            >>> new_job = offset( geometries = [geom1,geom2,...],
                                  offset_distance = 100,
                                  offset_unit = "esriMeters",
                                  offset_how = "esriGeometryOffsetRounded",
                                  bevel_ratio = 0,
                                  simplify_result = True
                                  spatial_ref = "wkid",
                                  future = True)

    """
    if gis is None:
        gis = arcgis.env.active_gis
    if isinstance(offset_unit, LengthUnits):
        offset_unit = offset_unit.value
    return gis._tools.geometry.offset(
        geometries,
        offset_distance,
        offset_unit,
        offset_how,
        bevel_ratio,
        simplify_result,
        spatial_ref,
        future=future,
    )


def project(
    geometries: Union[list[Polygon], list[Polyline], list[MultiPoint], list[Point]],
    in_sr: Optional[Union[int, dict[str, Any]]],
    out_sr: Optional[Union[int, dict[str, Any]]],
    transformation: str = "",
    transform_forward: bool = False,
    gis: Optional[GIS] = None,
    future: bool = False,
):
    """
    The ``project`` function is performed on a :class:`~arcgis.geometry.Geometry` service resource.
    This function projects an array of input geometries from the input
    :class:`~arcgis.geometry.SpatialReference` to the output :class:`~arcgis.geometry.SpatialReference`

    ================  ===============================================================================
    **Keys**          **Description**
    ----------------  -------------------------------------------------------------------------------
    geometries        An array of :class:`~arcgis.geometry.Point`, :class:`~arcgis.geometry.MultiPoint`,
                      :class:`~arcgis.geometry.Polyline`, or :class:`~arcgis.geometry.Polygon` objects.
                      The structure of each geometry in the array is the
                      same as the structure of the JSON geometry objects returned by
                      the ArcGIS REST API.
    ----------------  -------------------------------------------------------------------------------
    in_sr             The well-known ID of the :class:`~arcgis.geometry.SpatialReference` or a spatial
                      reference JSON object for the input geometries.
    ----------------  -------------------------------------------------------------------------------
    out_sr            The well-known ID of the :class:`~arcgis.geometry.SpatialReference` or a
                      spatial reference JSON object for the output geometries.
    ----------------  -------------------------------------------------------------------------------
    transformations   The WKID or a JSON object specifying the
                      geographic transformation (gis,also known as datum transformation) to be applied to the projected
                      geometries.

                      .. note::
                        A transformation is needed only if the output :class:`~arcgis.geometry.SpatialReference`
                        contains a different geographic coordinate system than the input
                        spatial reference.

    ----------------  -------------------------------------------------------------------------------
    transformforward  A Boolean value indicating whether or not to transform forward. The forward or reverse direction
                      of transformation is implied in the name of the transformation. If
                      transformation is specified, a value for the ``transform_Forward``
                      parameter must also be specified. The default value is false.
    ----------------  -------------------------------------------------------------------------------
    future            Optional boolean. If True, a future object will be returned and the process
                      will not wait for the task to complete. The default is False, which means wait for results.
                      If setting future to True there is a limitation of 6500 geometries that can be processed in one call.
    ================  ===============================================================================

    :returns:
        A list of :class:`~arcgis.geometry.Geometry` objects in the ``out_sr`` coordinate system, or
        a `GeometryJob` object. If ``future = True``,
        then the result is a :class:`~concurrent.futures.Future` object. Call ``result()`` to get the response.

    .. code-block:: python

        #Usage Example

        >>> result = project(geometries = [{"x": -17568824.55, "y": 2428377.35}, {"x": -17568456.88, "y": 2428431.352}],
                             in_sr = 3857,
                             out_sr = 4326)
            [{"x": -157.82343617279275, "y": 21.305781607280093}, {"x": -157.8201333369876, "y": 21.306233559873714}]
    """
    if gis is None:
        gis = arcgis.env.active_gis
    return gis._tools.geometry.project(
        geometries,
        in_sr,
        out_sr,
        transformation,
        transform_forward,
        future=future,
    )


def relation(
    geometries1: list[Geometry],
    geometries2: list[Geometry],
    spatial_ref: Optional[Union[int, dict[str, Any]]],
    spatial_relation: str = "esriGeometryRelationIntersection",
    relation_param: str = "",
    gis: Optional[GIS] = None,
    future: bool = False,
):
    """
    The ``relation`` function is performed on a :class:`~arcgis.geometry.Geometry` service resource.
    This function determines the pairs of geometries from the input
    geometry arrays that participate in the specified spatial relation.
    Both arrays are assumed to be in the spatial reference specified by
    ``spatial_ref``, which is a required parameter. Geometry types cannot be mixed
    within an array.

    .. note::
        The relations are evaluated in 2D. In other words, `z` coordinates are not used.

    ================  ===============================================================================
    **Keys**          **Description**
    ----------------  -------------------------------------------------------------------------------
    geometry1         The first array of :class:`~arcgis.geometry.Geometry` objects to compute relations.
    ----------------  -------------------------------------------------------------------------------
    geometry2         The second array of :class:`~arcgis.geometry.Geometry` objects to compute relations.
    ----------------  -------------------------------------------------------------------------------
    relation_param    The Shape Comparison Language string to be evaluated.
    ----------------  -------------------------------------------------------------------------------
    spatial_relation  The spatial relationship to be tested between the two input geometry arrays.
                      Values: `esriGeometryRelationCross | esriGeometryRelationDisjoint |
                      esriGeometryRelationIn | esriGeometryRelationInteriorIntersection |
                      esriGeometryRelationIntersection | esriGeometryRelationLineCoincidence |
                      esriGeometryRelationLineTouch | esriGeometryRelationOverlap |
                      esriGeometryRelationPointTouch | esriGeometryRelationTouch |
                      esriGeometryRelationWithin | esriGeometryRelationRelation`
    ----------------  -------------------------------------------------------------------------------
    spatial_ref       A :class:`~arcgis.geometry.SpatialReference` of the input geometries Well-Known ID or JSON object
    ----------------  -------------------------------------------------------------------------------
    future            Optional boolean. If True, a future object will be returned and the process
                      will not wait for the task to complete. The default is False, which means wait for results.
    ================  ===============================================================================


    :returns:
        A JSON dict of geometryNIndex between two lists of geometries, or a `GeometryJob` object. If ``future = True``,
        then the result is a :class:`~concurrent.futures.Future` object. Call ``result()`` to get the response.

            >>> new_res = relation(geometry1 = [geom1,geom2,...],
                                   geometry2 = [geom21,geom22,..],
                                   relation_param = "relationParameter",
                                   spatial_relation = "esriGeometryRelationPointTouch"
                                   spatial_ref = "wkid",
                                   future = False)
            >>> new_res
                {'relations': [{'geometry1Index': 0, 'geometry2Index': 0}]}
    """
    if gis is None:
        gis = arcgis.env.active_gis
    return gis._tools.geometry.relation(
        geometries1,
        geometries2,
        spatial_ref,
        spatial_relation,
        relation_param,
        future=future,
    )


def reshape(
    spatial_ref: Optional[Union[int, dict[str, Any]]],
    target: Union[Polyline, Polygon],
    reshaper: Polyline,
    gis: Optional[GIS] = None,
    future: bool = False,
):
    """
    The ``reshape`` function is performed on a :class:`~arcgis.geometry.Geometry` service resource.
    It reshapes a :class:`~arcgis.geometry.Polyline` or :class:`~arcgis.geometry.Polygon` feature by constructing a
    polyline over the feature. The feature takes the shape of the
    `reshaper` polyline from the first place the `reshaper` intersects the
    feature to the last.

    ================  ===============================================================================
    **Keys**          **Description**
    ----------------  -------------------------------------------------------------------------------
    target            The :class:`~arcgis.geometry.Polyline` or :class:`~arcgis.geometry.Polygon` to be reshaped.
    ----------------  -------------------------------------------------------------------------------
    reshaper          The single-part :class:`~arcgis.geometry.Polyline` that does the reshaping.
    ----------------  -------------------------------------------------------------------------------
    spatial_ref       A :class:`~arcgis.geometry.SpatialReference` of the input geometries Well-Known ID or a JSON
                      object for the input geometry
    ----------------  -------------------------------------------------------------------------------
    future            Optional boolean. If True, a future object will be returned and the process
                      will not wait for the task to complete. The default is False, which means wait for results.
    ================  ===============================================================================

    :returns:
        A reshaped :class:`~arcgis.geometry.Polyline` or :class:`~arcgis.geometry.Polygon` object, or
        a `GeometryJob` object. If ``future = True``,
        then the result is a :class:`~concurrent.futures.Future` object. Call ``result()`` to get the response.
    """
    if gis is None:
        gis = arcgis.env.active_gis
    return gis._tools.geometry.reshape(spatial_ref, target, reshaper, future=future)


def simplify(
    spatial_ref: Optional[Union[int, dict[str, Any]]],
    geometries: Union[list[Polygon], list[Polyline], list[MultiPoint], list[Point]],
    gis: Optional[GIS] = None,
    future: bool = False,
):
    """
    The ``simplify`` function is performed on a :class:`~arcgis.geometry.Geometry` service resource.
    ``simplify`` permanently alters the input geometry so that the geometry
    becomes topologically consistent. This resource applies the ArcGIS
    ``simplify`` function to each geometry in the input array.

    ================  ===============================================================================
    **Keys**          **Description**
    ----------------  -------------------------------------------------------------------------------
    geometries        An array of :class:`~arcgis.geometry.Point`, :class:`~arcgis.geometry.MultiPoint`,
                      :class:`~arcgis.geometry.Polyline`, or :class:`~arcgis.geometry.Polygon` objects.
                      The structure of each geometry in the array is the
                      same as the structure of the JSON geometry objects returned by
                      the ArcGIS REST API.
    ----------------  -------------------------------------------------------------------------------
    spatial_ref       A :class:`~arcgis.geometry.SpatialReference` of the input geometries Well-Known ID or JSON object
    ----------------  -------------------------------------------------------------------------------
    future            Optional boolean. If True, a future object will be returned and the process
                      will not wait for the task to complete. The default is False, which means wait for results.
                      If setting future to True there is a limitation of 6500 geometries that can be processed in one call.
    ================  ===============================================================================

    :returns:
        An array of :class:`~arcgis.geometry.Geometry` objects, or a `GeometryJob` object. If ``future = True``,
        then the result is a :class:`~concurrent.futures.Future` object. Call ``result()`` to get the response.
    """
    if gis is None:
        gis = arcgis.env.active_gis
    return gis._tools.geometry.simplify(spatial_ref, geometries, future=future)


def to_geo_coordinate_string(
    spatial_ref: Optional[Union[int, dict[str, Any]]],
    coordinates: json,
    conversion_type: str,
    conversion_mode: str = "mgrsDefault",
    num_of_digits: Optional[int] = None,
    rounding: bool = True,
    add_spaces: bool = True,
    gis: Optional[GIS] = None,
    future: bool = False,
):
    """
    The ``to_geo_coordinate_string`` function is performed on a :class:`~arcgis.geometry.Geometry`
    service resource. The function converts an array of
    xy-coordinates into well-known strings based on the conversion type
    and :class:`~arcgis.geometry.SpatialReference` supplied by the :class:`~arcgis.gis.User`. Optional parameters are
    available for some conversion types. See :attr:`~arcgis.geometry.functions.from_geo_coordinate_strings` for more
    information on the opposite conversion.

    .. note::
        If an optional parameter is not applicable for a particular conversion type, but a
        value is supplied for that parameter, the value will be ignored.


    ================  ===============================================================================
    **Keys**          **Description**
    ----------------  -------------------------------------------------------------------------------
    spatial_ref       A :class:`~arcgis.geometry.SpatialReference` of the input geometries Well-Known ID or JSON object
    ----------------  -------------------------------------------------------------------------------
    coordinates       An array of xy-coordinates in JSON format to be converted. Syntax: [[x1,y2],...[xN,yN]]
    ----------------  -------------------------------------------------------------------------------
    conversion-type   The conversion type of the input strings.

                      .. note::
                        Valid conversion types are:
                        `MGRS` - Military Grid Reference System
                        `USNG` - United States National Grid
                        `UTM` - Universal Transverse Mercator
                        `GeoRef` - World Geographic Reference System
                        `GARS` - Global Area Reference System
                        `DMS` - Degree Minute Second
                        `DDM` - Degree Decimal Minute
                        `DD` - Decimal Degree
    ----------------  -------------------------------------------------------------------------------
    conversion_mode   Conversion options for MGRS, UTM and GARS conversion types.

                      .. note::
                        Valid conversion modes for MGRS are:
                        `mgrsDefault` - Default. Uses the spheroid from the given spatial reference.

                        `mgrsNewStyle` - Treats all spheroids as new, like WGS 1984. The 80 degree longitude falls into Zone 60.

                        `mgrsOldStyle` - Treats all spheroids as old, like Bessel 1841. The 180 degree longitude falls into Zone 60.

                        `mgrsNewWith180InZone01` - Same as mgrsNewStyle except the 180 degree longitude falls into Zone 01

                        `mgrsOldWith180InZone01` - Same as mgrsOldStyle except the 180 degree longitude falls into Zone 01

                      .. note::
                        Valid conversion modes for UTM are:
                        `utmDefault` - Default. No options.
                        `utmNorthSouth` - Uses north/south latitude indicators instead of
                        `zone numbers` - Non-standard. Default is recommended
    ----------------  -------------------------------------------------------------------------------
    num_of_digits     The number of digits to output for each of the numerical portions in the string. The default
                      value for ``num_of_digits`` varies depending on ``conversion_type``.
    ----------------  -------------------------------------------------------------------------------
    rounding          If ``True``, then numeric portions of the string are rounded to the nearest whole magnitude as
                      specified by
                      num_of_digits. Otherwise, numeric portions of the string are
                      truncated. The rounding parameter applies only to conversion
                      types `MGRS`, `USNG` and `GeoRef`. The default value is ``True``.
    ----------------  -------------------------------------------------------------------------------
    addSpaces         If ``True``, then spaces are added between components of the string. The ``addSpaces`` parameter
                      applies only to conversion types `MGRS`, `USNG` and `UTM`. The default value for `MGRS` is
                      ``False``, while the default value for both `USNG` and `UTM` is ``True``.
    ----------------  -------------------------------------------------------------------------------
    future            Optional boolean. If True, a future object will be returned and the process
                      will not wait for the task to complete. The default is False, which means wait for results.
    ================  ===============================================================================

    :returns:
        An array of Strings, or a `GeometryJob` object. If ``future = True``,
        then the result is a :class:`~concurrent.futures.Future` object. Call ``result()`` to get the response.

        .. code-block:: python

            >>> strings = from_geo_coordinate_string(spatial_ref = "wkid",
                                                     coordinates = [[x1,y1], [x2,y2], [x3,y3]]
                                                     conversion_type = "MGRS",
                                                     conversion_mode = "mgrs_default",
                                                     future = False)
            >>> strings
                ["01N AA 66021 00000","11S NT 00000 62155", "31U BT 94071 65288"]
    """
    if gis is None:
        gis = arcgis.env.active_gis
    return gis._tools.geometry.to_geo_coordinate_string(
        spatial_ref,
        coordinates,
        conversion_type,
        conversion_mode,
        num_of_digits,
        rounding,
        add_spaces,
        future=future,
    )


def trim_extend(
    spatial_ref: Optional[Union[int, dict[str, Any]]],
    polylines: list[Polyline],
    trim_extend_to: Polyline,
    extend_how: int = 0,
    gis: Optional[GIS] = None,
    future: bool = False,
):
    """
    The ``trim_extend`` function is performed on a :class:`~arcgis.geometry.Geometry` service
    resource. This function trims or extends each :class:`~arcgis.geometry.Polyline` specified
    in the input array, using the user-specified guide polylines.

    .. note::
        When trimming features, the part to the left of the oriented cutting
        line is preserved in the output, and the other part is discarded.
        An empty :class:`~arcgis.geometry.Polyline` is added to the output array if the corresponding
        input polyline is neither cut nor extended.

    ================  ===============================================================================
    **Keys**          **Description**
    ----------------  -------------------------------------------------------------------------------
    polylines         An array of :class:`~arcgis.geometry.Polyline` objects to trim or extend
    ----------------  -------------------------------------------------------------------------------
    trim_extend_to    A :class:`~arcgis.geometry.Polyline` that is used as a guide for trimming or
                      extending input polylines.
    ----------------  -------------------------------------------------------------------------------
    extend_how        A flag that is used along with the trimExtend function.

                      ``0`` - By default, an extension considers both ends of a path. The
                      old ends remain, and new points are added to the extended ends.
                      The new points have attributes that are extrapolated from adjacent existing segments.

                      ``1`` - If an extension is performed at an end, relocate the end
                      point to the new position instead of leaving the old point and
                      adding a new point at the new position.

                      ``2`` - If an extension is performed at an end, do not extrapolate
                      the end-segment's attributes for the new point. Instead, make
                      its attributes the same as the current end. Incompatible with `esriNoAttributes`.

                      ``4`` - If an extension is performed at an end, do not extrapolate
                      the end-segment's attributes for the new point. Instead, make
                      its attributes empty. Incompatible with esriKeepAttributes.

                      ``8`` - Do not extend the 'from' end of any path.

                      ``16`` - Do not extend the 'to' end of any path.
    ----------------  -------------------------------------------------------------------------------
    spatial_ref       A :class:`~arcgis.geometry.SpatialReference` of the input geometries Well-Known ID or JSON object
    ----------------  -------------------------------------------------------------------------------
    future            Optional boolean. If True, a future object will be returned and the process
                      will not wait for the task to complete. The default is False, which means wait for results.
    ================  ===============================================================================

    :returns:
        An array of :class:`~arcgis.geometry.Polyline` objects, or a `GeometryJob` object. If ``future = True``,
        then the result is a :class:`~concurrent.futures.Future` object. Call ``result()`` to get the response.

    .. code-block:: python

            >>> polylines_arr = trim_extends(polylines = [polyline1,polyline2, ...],
                                             trim_extend_to = polyline_trimmer
                                             extend_how = 2,
                                             spatial_ref = "wkid",
                                             future = False)
            >>> polyline_arr
                [polyline1, polyline2,...]
    """
    if gis is None:
        gis = arcgis.env.active_gis
    return gis._tools.geometry.trim_extend(
        spatial_ref, polylines, trim_extend_to, extend_how, future=future
    )


def union(
    geometries: Union[list[Polygon], list[Polyline], list[MultiPoint], list[Point]],
    spatial_ref: Optional[Union[str, dict[str:str]]] = None,
    gis: Optional[GIS] = None,
    future: bool = False,
):
    """
    The ``union`` function is performed on a :class:`~arcgis.geometry.Geometry` service resource.
    This function constructs the set-theoretic union of the geometries
    in the input array.

    .. note::
        All inputs must be of the same type.

    ================  ===============================================================================
    **Keys**          **Description**
    ----------------  -------------------------------------------------------------------------------
    geometries        Required. An array of :class:`~arcgis.geometry.Point`,
                      :class:`~arcgis.geometry.MultiPoint`, :class:`~arcgis.geometry.Polyline`,
                      or :class:`~arcgis.geometry.Polygon` objects.
                      The structure of each geometry in the array is the
                      same as the structure of the JSON geometry objects returned by
                      the ArcGIS REST API.
    ----------------  -------------------------------------------------------------------------------
    spatial_ref       An optional String or JSON Dict representing the wkid to be used. The default is the
                      spatial reference found in the geometry or, if None found, then "4326".

                      Example: "4326" or {"wkid":"4326"}
    ----------------  -------------------------------------------------------------------------------
    future            Optional boolean. If True, a future object will be returned and the process
                      will not wait for the task to complete. The default is False, which means wait for results.
                      If setting future to True there is a limitation of 6500 geometries that can be processed in one call.
    ================  ===============================================================================

    :returns:
        The set-theoretic union of the :class:`~arcgis.geometry.Geometry` objects, or a `GeometryJob` object. If ``future = True``,
        then the result is a :class:`~concurrent.futures.Future` object. Call ``result()`` to get the response.
    """
    if gis is None:
        gis = arcgis.env.active_gis
    if spatial_ref is None:
        spatial_ref = [
            geom.spatialReference
            for geom in geometries
            if "spatialReference" in geom and geom.spatialReference is not None
        ]
        spatial_ref = spatial_ref[0] if len(spatial_ref) > 0 else "4326"
    if isinstance(spatial_ref, dict):
        spatial_ref = spatial_ref["wkid"]
    return gis._tools.geometry.union(spatial_ref, geometries, future=future)



from .rss import RSS
from .http_poller import HttpPoller
from .http_receiver import HttpReceiver
from .http_simulator import HttpSimulator
from .aws_iot import AWSIoT
from .azure_event_hub import AzureEventHub
from .azure_service_bus import AzureServiceBus
from .cisco_edge_intelligence import CiscoEdgeIntelligence
from .feature_layer import FeatureLayer
from .stream_layer import StreamLayer
from .geotab import Geotab
from .kafka import Kafka
from .mqtt import MQTT
from .rabbit_mq import RabbitMQ
from .verizon_connect_reveal import VerizonConnectReveal
from .web_socket import WebSocket
from .kafka_authentication_type import (
    NoAuth,
    SASLPlain,
    SaslScramSha256,
    SaslScramSha512,
)
from .geometry import XYZGeometry, SingleFieldGeometry
from .time import TimeInstant, TimeInterval
from .run_interval import RunInterval



# ------------------------------------------------------------------------
# Deformable DETR
# Copyright (c) 2020 SenseTime. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 [see LICENSE for details]
# ------------------------------------------------------------------------
# Modified from DETR (https://github.com/facebookresearch/detr)
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
# ------------------------------------------------------------------------

"""
Modules to compute the matching cost and solve the corresponding LSAP.
"""
import torch
from scipy.optimize import linear_sum_assignment
from torch import nn
from arcgis.learn._utils.coco_detection_utils import (
    box_cxcywh_to_xyxy,
    generalized_box_iou,
)


class HungarianMatcher(nn.Module):
    """This class computes an assignment between the targets and the predictions of the network

    For efficiency reasons, the targets don't include the no_object. Because of this, in general,
    there are more predictions than targets. In this case, we do a 1-to-1 matching of the best predictions,
    while the others are un-matched (and thus treated as non-objects).
    """

    def __init__(
        self, cost_class: float = 1, cost_bbox: float = 1, cost_giou: float = 1
    ):
        """Creates the matcher

        Params:
            cost_class: This is the relative weight of the classification error in the matching cost
            cost_bbox: This is the relative weight of the L1 error of the bounding box coordinates in the matching cost
            cost_giou: This is the relative weight of the giou loss of the bounding box in the matching cost
        """
        super().__init__()
        self.cost_class = cost_class
        self.cost_bbox = cost_bbox
        self.cost_giou = cost_giou
        assert (
            cost_class != 0 or cost_bbox != 0 or cost_giou != 0
        ), "all costs cant be 0"

    def forward(self, outputs, targets):
        """Performs the matching

        Params:
            outputs: This is a dict that contains at least these entries:
                 "pred_logits": Tensor of dim [batch_size, num_queries, num_classes] with the classification logits
                 "pred_boxes": Tensor of dim [batch_size, num_queries, 4] with the predicted box coordinates

            targets: This is a list of targets (len(targets) = batch_size), where each target is a dict containing:
                 "labels": Tensor of dim [num_target_boxes] (where num_target_boxes is the number of ground-truth
                           objects in the target) containing the class labels
                 "boxes": Tensor of dim [num_target_boxes, 4] containing the target box coordinates

        Returns:
            A list of size batch_size, containing tuples of (index_i, index_j) where:
                - index_i is the indices of the selected predictions (in order)
                - index_j is the indices of the corresponding selected targets (in order)
            For each batch element, it holds:
                len(index_i) = len(index_j) = min(num_queries, num_target_boxes)
        """
        with torch.no_grad():
            bs, num_queries = outputs["pred_logits"].shape[:2]

            # We flatten to compute the cost matrices in a batch
            out_prob = outputs["pred_logits"].flatten(0, 1).sigmoid()
            out_bbox = outputs["pred_boxes"].flatten(
                0, 1
            )  # [batch_size * num_queries, 4]

            # Also concat the target labels and boxes
            tgt_ids = torch.cat([v["labels"] for v in targets])
            tgt_bbox = torch.cat([v["boxes"] for v in targets])

            # Compute the classification cost.
            alpha = 0.25
            gamma = 2.0
            neg_cost_class = (
                (1 - alpha) * (out_prob**gamma) * (-(1 - out_prob + 1e-8).log())
            )
            pos_cost_class = (
                alpha * ((1 - out_prob) ** gamma) * (-(out_prob + 1e-8).log())
            )
            cost_class = pos_cost_class[:, tgt_ids] - neg_cost_class[:, tgt_ids]

            # Compute the L1 cost between boxes
            cost_bbox = torch.cdist(out_bbox, tgt_bbox, p=1)
            # print(out_bbox)

            # Compute the giou cost betwen boxes
            cost_giou = -generalized_box_iou(
                box_cxcywh_to_xyxy(out_bbox), box_cxcywh_to_xyxy(tgt_bbox)
            )

            # Final cost matrix
            C = (
                self.cost_bbox * cost_bbox
                + self.cost_class * cost_class
                + self.cost_giou * cost_giou
            )
            C = C.view(bs, num_queries, -1).cpu()

            sizes = [len(v["boxes"]) for v in targets]
            indices = [
                linear_sum_assignment(c[i]) for i, c in enumerate(C.split(sizes, -1))
            ]
            return [
                (
                    torch.as_tensor(i, dtype=torch.int64),
                    torch.as_tensor(j, dtype=torch.int64),
                )
                for i, j in indices
            ]


def build_matcher(args):
    return HungarianMatcher(
        cost_class=args.set_cost_class,
        cost_bbox=args.set_cost_bbox,
        cost_giou=args.set_cost_giou,
    )



# Copyright 2018-2019 Open-MMLab.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# error in show results
# ga_faster_x101_64x4d_fpn_1x_coco.py, box AP=43.9

_base_ = "./_base_/models/faster_rcnn_r50_fpn.py"
model = dict(
    pretrained="open-mmlab://resnext101_64x4d",
    backbone=dict(
        type="ResNeXt",
        depth=101,
        groups=64,
        base_width=4,
        num_stages=4,
        out_indices=(0, 1, 2, 3),
        frozen_stages=1,
        norm_cfg=dict(type="BN", requires_grad=True),
        style="pytorch",
    ),
    rpn_head=dict(
        _delete_=True,
        type="GARPNHead",
        in_channels=256,
        feat_channels=256,
        approx_anchor_generator=dict(
            type="AnchorGenerator",
            octave_base_scale=8,
            scales_per_octave=3,
            ratios=[0.5, 1.0, 2.0],
            strides=[4, 8, 16, 32, 64],
        ),
        square_anchor_generator=dict(
            type="AnchorGenerator", ratios=[1.0], scales=[8], strides=[4, 8, 16, 32, 64]
        ),
        anchor_coder=dict(
            type="DeltaXYWHBBoxCoder",
            target_means=[0.0, 0.0, 0.0, 0.0],
            target_stds=[0.07, 0.07, 0.14, 0.14],
        ),
        bbox_coder=dict(
            type="DeltaXYWHBBoxCoder",
            target_means=[0.0, 0.0, 0.0, 0.0],
            target_stds=[0.07, 0.07, 0.11, 0.11],
        ),
        loc_filter_thr=0.01,
        loss_loc=dict(
            type="FocalLoss", use_sigmoid=True, gamma=2.0, alpha=0.25, loss_weight=1.0
        ),
        loss_shape=dict(type="BoundedIoULoss", beta=0.2, loss_weight=1.0),
        loss_cls=dict(type="CrossEntropyLoss", use_sigmoid=True, loss_weight=1.0),
        loss_bbox=dict(type="SmoothL1Loss", beta=1.0, loss_weight=1.0),
    ),
    roi_head=dict(bbox_head=dict(bbox_coder=dict(target_stds=[0.05, 0.05, 0.1, 0.1]))),
    # model training and testing settings
    train_cfg=dict(
        rpn=dict(
            ga_assigner=dict(
                type="ApproxMaxIoUAssigner",
                pos_iou_thr=0.7,
                neg_iou_thr=0.3,
                min_pos_iou=0.3,
                ignore_iof_thr=-1,
            ),
            ga_sampler=dict(
                type="RandomSampler",
                num=256,
                pos_fraction=0.5,
                neg_pos_ub=-1,
                add_gt_as_proposals=False,
            ),
            allowed_border=-1,
            center_ratio=0.2,
            ignore_ratio=0.5,
        ),
        rpn_proposal=dict(max_num=300),
        rcnn=dict(
            assigner=dict(pos_iou_thr=0.6, neg_iou_thr=0.6, min_pos_iou=0.6),
            sampler=dict(type="RandomSampler", num=256),
        ),
    ),
    test_cfg=dict(rpn=dict(max_num=300), rcnn=dict(score_thr=1e-3)),
)

checkpoint = "http://download.openmmlab.com/mmdetection/v2.0/guided_anchoring/ga_faster_x101_64x4d_fpn_1x_coco/ga_faster_x101_64x4d_fpn_1x_coco_20200215-0fa7bde7.pth"



class DaskSpatialIndex:
    """

    A spatial index is a type of extended index that allows you to index a
    spatial column. A spatial column is a table column that contains data of a
    spatial data type.

    Spatial indexes help to improve spatial query performance on a dataframe.
    Identifying a feature, selecting features, and joining data all have better
    performace when using spatial indexing.


    ====================     ==================================================
    Arguement                Description
    --------------------     --------------------------------------------------
    indexes                  A collection of spatial index objects.
    ====================     ==================================================


    """

    _indexes = None

    # ----------------------------------------------------------------------
    def __init__(self, indexes: tuple):
        """initializer"""
        self._indexes = indexes

    def __str__(self):
        return "<Dask QuadTree Index>"

    # ----------------------------------------------------------------------
    def intersect(self, bbox: tuple) -> list:
        """
        Returns the spatial features that intersect the bbox

        :bbox: tuple - (xmin,ymin,xmax,ymax)

        :returns: list
        """
        res = []
        [res.extend(index.intersect(bbox)) for index in self._indexes]
        return res
